##
# File:    MyDdSqlGen.py
# Author:  J. Westbrook
# Date:    31-Jan-2012
# Version: 0.001 Initial version
#
# Updates:
# 27-Jan-2012 Jdw Refactored from MyDbUtil to isolate portable SQL generators.
#  1-Feb-2012 Jdw Add export/import methods
# 11-Apr-2014 jdw add template methods with attribute Id inputs.
# 25-May-2015 jdw complete the coding of the contraint generator class
# 28-May-2015 jdw adjust terminology in api method names and internal vars -
# 16-Jun-2015 jdw generalized the addition of a condition group addGroupValueConditionList
#  2-Oct-2017 jdw py3 compatibility use zip_longest
#  2-Oct-2017 jdw fix obvious error with missing len() in addKeyAttributeEquiJoinConditions()
# 30-Dec-2017 jdw add crate specific SQL generators
#  5-Jan-2018 jdw add default replication factor for crate -
# 20-Jun-2018 jdw adjustments for dynamic schema generation
#  7-Jul-2018 jdw update for new schema def prototypes
#
#
##
"""
A collection of classes to generate SQL commands to perform queries and schema construction.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import copy
import itertools
import logging

# from operator import itemgetter, attrgetter
try:
    from itertools import zip_longest as zip_longest
except Exception:
    from itertools import izip_longest as zip_longest

logger = logging.getLogger(__name__)
#


class SqlGenAdmin(object):

    """ Builds SQL commands to create table schema from a schema definition derived from class SchemaDefBase.

        Note:
    """

    def __init__(self, verbose=False, serverType="mysql"):

        self.__verbose = verbose
        self.__serverType = str(serverType).lower()

    def truncateTableSQL(self, databaseName, tableName):
        """ Return the SQL string require to truncate (remove all rows) from the input table.
        """
        return "TRUNCATE TABLE %s.%s; " % (databaseName, tableName)

    def idUpdateTemplateSQL(self, databaseName, tableDefObj, updateAttributeIdList=[], conditionAttributeIdList=[]):
        """ Return the SQL string template for updating the input attributes into the named table subject
            to the constraining attributes.

            The string provides formatting placeholders for updated values  as well as for constraining values.

            The input table object is used to adjust the value quoting in the returned template string.

        """
        tableName = tableDefObj.getName()
        fL = []
        for atId in updateAttributeIdList:
            if tableDefObj.isAttributeStringType(atId):
                fL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
            elif tableDefObj.isAttributeFloatType(atId):
                fL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
            elif tableDefObj.isAttributeIntegerType(atId):
                fL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
            else:
                fL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
        #
        if len(conditionAttributeIdList) > 0:
            cL = []
            for atId in conditionAttributeIdList:
                if tableDefObj.isAttributeStringType(atId):
                    cL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
                else:
                    cL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
            tS = "UPDATE %s.%s SET %s WHERE (%s);" % (databaseName, tableName, ','.join(fL), ','.join(cL))
        else:
            tS = "UPDATE %s.%s SET %s;" % (databaseName, tableName, ','.join(fL))
        #
        return tS

    def idInsertTemplateSQL(self, databaseName, tableDefObj, insertAttributeIdList=[]):
        """  Return the SQL string template for inserting the input attributes into the named table.

            The string provides formatting placeholders for updated values  as well as for constraining values.

            The input table object is used to adjust the value quoting in the returned template string.

        """
        tableName = tableDefObj.getName()
        attributeNameList = []
        #
        fL = []
        for atId in insertAttributeIdList:
            attributeNameList.append(tableDefObj.getAttributeName(atId))
            if tableDefObj.isAttributeStringType(atId):
                fL.append('%s')
            elif tableDefObj.isAttributeFloatType(atId):
                fL.append('%s')
            elif tableDefObj.isAttributeIntegerType(atId):
                fL.append('%s')
            else:
                fL.append('%s')
        #
        #logger.debug(" %r %r %r %r" % (databaseName, tableName, attributeNameList, fL))
        tS = "INSERT INTO %s.%s (%s) VALUES (%s);" % (databaseName, tableName, ','.join(attributeNameList), ','.join(fL))
        #
        return tS

    def idDeleteTemplateSQL(self, databaseName, tableDefObj, conditionAttributeIdList=[]):
        """ Return the SQL string template for deleting records in the named table subject
            to the constraining attributes.

            The string provides formatting placeholders for constraining values.

            The input table object is used to adjust the value quoting in the returned template string.

        """
        tableName = tableDefObj.getName()
        #
        if len(conditionAttributeIdList) > 0:
            cL = []
            for atId in conditionAttributeIdList:
                if tableDefObj.isAttributeStringType(atId):
                    cL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')
                else:
                    cL.append(' %s=' % tableDefObj.getAttributeName(atId) + '%s')

            tS = "DELETE FROM  %s.%s WHERE (%s);" % (databaseName, tableName, ' AND '.join(cL))
        else:
            tS = "DELETE FROM  %s.%s;" % (databaseName, tableName)
        #
        return tS

    def insertTemplateSQL(self, databaseName, tableName, attributeNameList=[]):
        """ Return the SQL string template for inserting the input attributes into the named table.

            The string provides formatting placeholders for inserted values that are added when
            the SQL command is executed.

        """
        fL = []
        for v in attributeNameList:
            fL.append('%s')
        #
        tS = "INSERT INTO %s.%s (%s) VALUES (%s);" % (databaseName, tableName, ','.join(attributeNameList), ','.join(fL))
        return tS

    def insertTemplateSQLCrate(self, databaseName, tableName, attributeNameList=[]):
        """ Return the SQL string template for inserting the input attributes into the named table.

            The string provides formatting placeholders for inserted values that are added when
            the SQL command is executed.

                cursor.executemany("INSERT INTO locations (name, date, kind, position) VALUES (?, ?, ?, ?)",
                ...                [('Cloverleaf', '2007-03-11', 'Quasar', 7),
                ...                 ('Old Faithful', '2007-03-11', 'Quasar', 7)])
                [{u'rowcount': 1}, {u'rowcount': 1}]
        """
        fL = []
        for v in attributeNameList:
            fL.append('?')
        #
        tS = "INSERT INTO %s.%s (%s) VALUES (%s);" % (databaseName, tableName, ','.join(attributeNameList), ','.join(fL))
        return tS

    def refreshTableSQLCrate(self, databaseName, tableName):
        """ Return the SQL string to refresh named table.

        """
        #
        tS = "REFRESH TABLE %s.%s;" % (databaseName, tableName)
        return tS

    def deleteTemplateSQL(self, databaseName, tableName, attributeNameList=[]):
        """ Return the SQL string template for deleting table records constrained by the input attributes.

            The string provides formatting placeholders for the constraining values.
             delete from <tableName>  where at1=%s and at2=%s

        """
        fL = []
        for v in attributeNameList:
            fL.append(' %s=' % v + '%s')
        #
        tS = "DELETE FROM %s.%s WHERE %s;" % (databaseName, tableName, ' AND '.join(fL))
        return tS

    def deleteFromListSQL(self, databaseName, tableName, attributeName, valueList, chunkSize=10):
        """ Return the SQL string for deleting table records for a list of string values of
            the input attribute.

             delete from <databaseName>.<tableName>  where attributeName  IN (v1,v2,v3);

        """
        sqlList = []
        chunkLists = self.__makeSubLists(chunkSize, valueList)
        for chunk in chunkLists:
            fL = ["'%s'" % v for v in chunk]
            sqlList.append("DELETE FROM %s.%s WHERE %s IN (%s); " % (databaseName, tableName, attributeName, ','.join(fL)))

        return sqlList

    def __makeSubLists(self, n, iterable):
        args = [iter(iterable)] * n
        return ([e for e in t if e is not None] for t in zip_longest(*args))

    def createDatabaseSQL(self, databaseName):
        """ Return a list of strings containing the SQL to drop and recreate the input database.

           DROP DATABASE IF EXISTS <databaseName>;
           CREATE DATABASE <databaseName>;
        """
        oL = []
        oL.append("DROP DATABASE IF EXISTS %s;" % databaseName)
        oL.append("CREATE DATABASE %s;" % databaseName)
        return oL

    def removeDatabaseSQL(self, databaseName):
        """ Return a list of strings containing the SQL to drop and recreate the input database.

           DROP DATABASE IF EXISTS <databaseName>;
           CREATE DATABASE <databaseName>;
        """
        oL = []
        oL.append("DROP DATABASE IF EXISTS %s;" % databaseName)
        return oL

    def createTableSQL(self, databaseName, tableDefObj):
        """ Return a list of strings containing the SQL commands to create the table and indices
            described by the input table definition.

        """
        oL = []
        if self.__serverType == "cratedb":
            oL.extend(self.__dropTableCrate(databaseName, tableDefObj.getName()))
            oL.extend(self.__createTableCrate(databaseName, tableDefObj))
            # oL.extend(self.__createTableIndices(tableDefObj))
        elif self.__serverType == 'cockroachdb':
            oL.extend(self.__setDatabase(databaseName))
            oL.extend(self.__dropTable(tableDefObj.getName()))
            oL.extend(self.__createTableCockroach(tableDefObj))
            oL.extend(self.__createTableIndices(tableDefObj))
        else:
            oL.extend(self.__setDatabase(databaseName))
            oL.extend(self.__dropTable(tableDefObj.getName()))
            oL.extend(self.__createTable(tableDefObj))
            oL.extend(self.__createTableIndices(tableDefObj))
        return oL

    def dropTableSQL(self, databaseName, tableDefObj):
        oL = []
        if self.__serverType == "cratedb":
            oL.extend(self.__dropTableCrate(databaseName, tableDefObj.getName()))
        else:
            oL.extend(self.__setDatabase(databaseName))
            oL.extend(self.__dropTable(tableDefObj.getName()))
        return oL

    def __setDatabase(self, databaseName):
        """Return a list of strings containing database connection SQL command for the input database

           USE <databaseName>;
        """
        return ["USE %s;" % databaseName]

    def __dropTable(self, tableName):
        """ Return a list of strings containing the SQL DROP TABLE command for the input table:

           DROP TABLE IF EXISTS <tableName>;
        """
        return ["DROP TABLE IF EXISTS %s;" % tableName]

    def __createTable(self, tableDefObj):
        """ Return a list of strings containing the SQL command to create the table described in
            input table schema definition object.

        """
        oL = []
        pkL = []
        #
        attributeIdList = tableDefObj.getAttributeIdList()
        #
        oL.append("CREATE TABLE %s (" % tableDefObj.getName())
        for ii, attributeId in enumerate(attributeIdList):
            #
            name = tableDefObj.getAttributeName(attributeId)

            sqlType = str(tableDefObj.getAttributeType(attributeId)).upper()
            width = int(tableDefObj.getAttributeWidth(attributeId))
            precision = int(tableDefObj.getAttributePrecision(attributeId))
            notNull = "not null" if not tableDefObj.getAttributeNullable(attributeId) else "    null default null"
            if tableDefObj.getAttributeIsPrimaryKey(attributeId):
                pkL.append(name)
            #
            if ((sqlType == "CHAR") or (sqlType == "VARCHAR")):
                sW = "%-s(%d)" % (sqlType, width)
                tS = "%-40s %-16s  %s" % (name, sW, notNull)
            elif (sqlType.startswith("INT") or sqlType in ["INTEGER", "BIGINT", "SMALLINT"]):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif (sqlType in ["FLOAT", "REAL", "DOUBLE PRECISION"]):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif ((sqlType == "DATE") or (sqlType == "DATETIME")):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif ((sqlType == "TEXT") or (sqlType == "MEDIUMTEXT") or (sqlType == "LONGTEXT")):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif ((sqlType == "DECIMAL") or (sqlType == "NUMERIC")):
                sW = "%-s(%d,%d)" % (sqlType, width, precision)
                tS = "%-40s %-16s  %s" % (name, sW, notNull)
            else:
                pass
            #
            # if ii < len(attributeIdList) -1:
            #    oL.append(tS+",")
            # else:
            #    oL.append(tS)
            oL.append(tS + ",")

        if len(pkL) > 0:
            oL.append("PRIMARY KEY (%s)" % (','.join(pkL)))

        if (str(tableDefObj.getType()).upper() == "TRANSACTIONAL"):
            oL.append(") ENGINE InnoDB;")
        else:
            oL.append(") ENGINE MyISAM;")
        #
        # return this as list containing a single string command.
        return ['\n'.join(oL)]

    def __createTableCockroach(self, tableDefObj):
        """ Return a list of strings containing the SQL command to create the table described in
            input table schema definition object.

        """
        oL = []
        pkL = []
        #
        attributeIdList = tableDefObj.getAttributeIdList()
        #
        oL.append("CREATE TABLE %s (" % tableDefObj.getName())
        for ii, attributeId in enumerate(attributeIdList):
            #
            name = tableDefObj.getAttributeName(attributeId)

            sqlType = str(tableDefObj.getAttributeType(attributeId)).upper()
            width = int(tableDefObj.getAttributeWidth(attributeId))
            precision = int(tableDefObj.getAttributePrecision(attributeId))
            notNull = "not null" if not tableDefObj.getAttributeNullable(attributeId) else "    null default null"
            if tableDefObj.getAttributeIsPrimaryKey(attributeId):
                pkL.append(name)
            #
            if ((sqlType == "CHAR") or (sqlType == "VARCHAR")):
                sW = "%-s(%d)" % (sqlType, width)
                tS = "%-40s %-16s  %s" % (name, sW, notNull)
            elif (sqlType.startswith("INT") or sqlType in ["INTEGER", "BIGINT", "SMALLINT"]):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif (sqlType in ["FLOAT", "REAL", "DOUBLE PRECISION"]):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif (sqlType == "DATE"):
                tS = "%-40s %-16s  %s" % (name, sqlType, notNull)
            elif (sqlType == "DATETIME"):
                tS = "%-40s %-16s  %s" % (name, 'timestamp', notNull)
            elif ((sqlType == "TEXT") or (sqlType == "MEDIUMTEXT") or (sqlType == "LONGTEXT")):
                tS = "%-40s %-16s  %s" % (name, 'text', notNull)
            elif ((sqlType == "DECIMAL") or (sqlType == "NUMERIC")):
                sW = "%-s(%d,%d)" % (sqlType, width, precision)
                tS = "%-40s %-16s  %s" % (name, sW, notNull)
            else:
                pass
            #
            # if ii < len(attributeIdList) -1:
            #    oL.append(tS+",")
            # else:
            #    oL.append(tS)
            oL.append(tS + ",")

        if len(pkL) > 0:
            oL.append("PRIMARY KEY (%s)" % (','.join(pkL)))

        if False:
            if (str(tableDefObj.getType()).upper() == "TRANSACTIONAL"):
                oL.append(") ENGINE InnoDB;")
            else:
                oL.append(") ENGINE MyISAM;")
        else:
            oL.append(");")
        #
        # return this as list containing a single string command.
        return ['\n'.join(oL)]

    def __dropTableCrate(self, databaseName, tableName):
        """ Return a list of strings containing the SQL DROP TABLE command for the input table:

           DROP TABLE IF EXISTS <tableName>;
        """
        return ["DROP TABLE IF EXISTS %s.%s;" % (databaseName, tableName)]

    def __filterColumnName(self, name):
        rName = str(name).lower()
        if rName[0].isdigit():
            rName = "the_" + rName
        elif rName in ['offset']:
            rName = 'the_' + rName
        return rName

    def __createTableCrate(self, databaseName, tableDefObj):
        """ Return a list of strings containing the SQL command to create the table described in
            input table schema definition object. CrateDb variant

        """
        oL = []
        pkL = []
        #
        attributeIdList = tableDefObj.getAttributeIdList()
        #
        tableName = str(tableDefObj.getName()).lower()
        oL.append("CREATE TABLE %s.%s (" % (databaseName, tableName))
        for ii, attributeId in enumerate(attributeIdList):
            #
            name = self.__filterColumnName(tableDefObj.getAttributeName(attributeId))

            sqlType = str(tableDefObj.getAttributeType(attributeId)).upper()
            # width = int(tableDefObj.getAttributeWidth(attributeId))
            # precision = int(tableDefObj.getAttributePrecision(attributeId))
            notNull = "not null" if not tableDefObj.getAttributeNullable(attributeId) else " "

            if tableDefObj.getAttributeIsPrimaryKey(attributeId):
                pkL.append(name)
            #
            if ((sqlType == "CHAR") or (sqlType == "VARCHAR")):
                #sW = "%-s(%d)" % ("string", width)
                tS = "%-40s %-16s  %s" % (name, 'string', notNull)
            elif (sqlType.startswith("INT") or sqlType in ["INTEGER", "BIGINT", "SMALLINT"]):
                tS = "%-40s %-16s  %s" % (name, "long", notNull)
            elif (sqlType in ["FLOAT", "REAL", "DOUBLE PRECISION"]):
                tS = "%-40s %-16s  %s" % (name, "float", notNull)
            elif ((sqlType == "DATE") or (sqlType == "DATETIME")):
                tS = "%-40s %-16s  %s" % (name, "string", notNull)
            elif ((sqlType == "TEXT") or (sqlType == "MEDIUMTEXT") or (sqlType == "LONGTEXT")):
                tS = "%-40s %-16s  %s" % (name, "string", notNull)
            elif ((sqlType == "DECIMAL") or (sqlType == "NUMERIC")):
                tS = "%-40s %-16s  %s" % (name, "float", notNull)
                ##sW = "%-s(%d,%d)" % (sqlType, width, precision)
                ##tS = "%-40s %-16s  %s" % (name, sW, notNull)
            else:
                pass
            #
            # if ii < len(attributeIdList) -1:
            #    oL.append(tS+",")
            # else:
            #    oL.append(tS)
            oL.append(tS + ",")

        tL = self.__createTableIndicesCrate(tableDefObj)
        for t in tL:
            oL.append(t)
        if len(pkL) > 0:
            oL.append("PRIMARY KEY (%s)" % (','.join(pkL)))

        oL.append(") WITH (number_of_replicas='1-all');")

        # return this as list containing a single string command.
        return ['\n'.join(oL)]

    def __createTableIndices(self, tableDefObj):
        """ Return a list of strings containing the SQL command to create any indices described in
            input table schema definition object.

        """
        oL = []
        tableName = tableDefObj.getName()
        indexNameList = tableDefObj.getIndexNames()
        for indexName in indexNameList:
            indexType = tableDefObj.getIndexType(indexName)
            if str(indexType).upper() == "SEARCH":
                indexType = ""
            tL = []
            tL.append("CREATE %s INDEX %s on %s (" % (indexType, indexName, tableName))
            attributeIdList = tableDefObj.getIndexAttributeIdList(indexName)
            for ii, attributeId in enumerate(attributeIdList):
                name = tableDefObj.getAttributeName(attributeId)
                tS = "%-s" % name
                if ii < len(attributeIdList) - 1:
                    tL.append(tS + ",")
                else:
                    tL.append(tS)
            tL.append(");")
            oL.append(' '.join(tL))

        return oL

    def __createTableIndicesCrate(self, tableDefObj):
        """ Return a list of strings containing the SQL command to create any indices described in
            input table schema definition object.
            with (analyzer = 'english')
        """
        oL = []
        # tableName = tableDefObj.getName()
        indexNameList = tableDefObj.getIndexNames()
        for indexName in indexNameList:
            # indexType = tableDefObj.getIndexType(indexName)
            tL = []
            tL.append("INDEX %s USING FULLTEXT (" % (indexName))
            attributeIdList = tableDefObj.getIndexAttributeIdList(indexName)
            attributeIdListS = []
            for ii, attributeId in enumerate(attributeIdList):
                sqlType = str(tableDefObj.getAttributeType(attributeId)).upper()
                if sqlType in ['VARCHAR', 'CHAR', 'TEXT', 'DATE', 'DATETIME', 'MEDIUMTEXT', 'LONGTEXT']:
                    attributeIdListS.append(attributeId)
            #
            for ii, attributeId in enumerate(attributeIdListS):

                name = self.__filterColumnName(tableDefObj.getAttributeName(attributeId))
                tS = "%-s" % name
                if ii < len(attributeIdListS) - 1:
                    tL.append(tS + ",")
                else:
                    tL.append(tS)
            # tL.append(") with (analyzer = 'english'),")
            tL.append(") ,")
            oL.append(' '.join(tL))

        return oL

    def exportTable(self, databaseName, tableDefObj, exportPath, withDoubleQuotes=False):
        """
        """
        tableName = tableDefObj.getName()
        aNames = tableDefObj.getAttributeNameList()
        #
        oL = []
        oL.append("SELECT %s " % ','.join(aNames))
        oL.append(" INTO OUTFILE %s " % exportPath)
        oL.append("FIELDS TERMINATED BY '&##&\\t' ")
        if (withDoubleQuotes):
            oL.append(" OPTIONALLY ENCLOSED BY '\"' ")
        oL.append("LINES  TERMINATED BY '$##$\\n' ")
        oL.append("FROM %s.%s " % (databaseName, tableName))
        oL.append(";")
        return '\n'.join(oL)

    def importTable(self, databaseName, tableDefObj, importPath, withTruncate=False, withDoubleQuotes=False):
        """ Create the SQL commands to import data files stored in charactore delimited data files into the
            in put database and table.    Input data may be optionally enclosed in double quotes.

            An options is provied to  pre-truncate the table before loading.

            Return:  a string containing the SQL for the load command.
        """
        tableName = tableDefObj.getName()
        aNames = tableDefObj.getAttributeNameList()
        #
        oL = []
        if withTruncate:
            oL.append("TRUNCATE TABLE %s.%s; " % (databaseName, tableName))

        # oL.append("SET @@GLOBAL.local_infile = 1; ")
        oL.append("LOAD DATA LOCAL INFILE '%s' " % importPath)
        oL.append("INTO TABLE  %s.%s " % (databaseName, tableName))
        oL.append("FIELDS TERMINATED BY '&##&\\t' ")

        if (withDoubleQuotes):
            oL.append("OPTIONALLY ENCLOSED BY '\"' ")

        oL.append("LINES  TERMINATED BY '$##$\\n' ")
        oL.append(" (%s) " % ','.join(aNames))
        oL.append(";")
        return ' '.join(oL)


class SqlGenQuery(object):

    """ Builds an the SQL command string for a selection query.
    """

    def __init__(self, schemaDefObj, verbose=False):
        """  Input:

             schemaDef is instance of class derived from SchemaDefBase().
        """
        self.__schemaDefObj = schemaDefObj
        self.__verbose = verbose
        #
        self.__setup()

    def __setup(self):
        self.__databaseName = self.__schemaDefObj.getDatabaseName()
        self.__selectList = []
        self.__orderList = []
        self.__conditionObj = None
        self.__sortOrder = "DESC"
        self.__limitStart = None
        self.__limitLength = None
        #

    def setDatabase(self, databaseName):
        self.__databaseName = databaseName

    def clear(self):
        self.__setup()

    def addSelectLimit(self, rowStart=None, rowLength=None):
        try:
            self.__limitStart = int(rowStart)
            self.__limitLength = int(rowLength)
            return True
        except Exception:
            return False

    def addSelectAttributeId(self, attributeTuple=(None, None)):
        """ Add the input attribute to the current attribute select list.

            where attributeTuple contains (tableId,attributeId)

        """
        self.__selectList.append(attributeTuple)
        return True

    def setOrderBySortOrder(self, dir="ASC"):
        """  The default sort order applied to attributes in the ORDER BY clause. (ASC|DESC)
        """
        self.__sortOrder = dir

    def addOrderByAttributeId(self, attributeTuple=(None, None), sortFlag='DEFAULT'):
        """ Add the input attribute to the current orderBy list.

            where attributeTuple contains (tableId,attributeId)

        """
        sf = self.__sortOrder if sortFlag == 'DEFAULT' else sortFlag
        self.__orderList.append((attributeTuple, sf))
        return True

    def setCondition(self, conditionObj):
        """ Set an instance of the condition object from the SqlGenCondition() class.

        """
        self.__conditionObj = conditionObj
        return True

    def getSql(self):
        """
        """
        return self.__makeSql()

    def __makeSql(self):
        """  Builds SQL string for the query from the current list of attributes, list of
             ORDER BY attributes and the constrainObj.
        """
        #
        if len(self.__selectList) == 0:
            return None
        #
        # Attribute names from select list -
        #
        aNames = [self.__schemaDefObj.getQualifiedAttributeName(aTup) for aTup in self.__selectList]
        #
        # Table Id's from the select list -
        #
        tIds = [aTup[0] for aTup in self.__selectList]
        #
        conditionSql = None
        if self.__conditionObj is not None:
            conditionSql = self.__conditionObj.getSql()
            tIds.extend(self.__conditionObj.getTableIdList())
        #
        oNames = []
        if len(self.__orderList) > 0:
            oNames = [self.__schemaDefObj.getQualifiedAttributeName(aTup) + " " + sortFlag for aTup, sortFlag in self.__orderList]
        #
        tIds = list(set(tIds))
        tNames = [self.__databaseName + '.' + self.__schemaDefObj.getSchemaName(tId) for tId in tIds]
        #
        oL = []
        oL.append("SELECT %s " % ','.join(aNames))
        oL.append(" FROM %s " % ','.join(tNames))
        if conditionSql is not None and len(conditionSql) > 0:
            oL.append(" WHERE %s " % conditionSql)
        #
        if len(oNames) > 0:
            oL.append(" ORDER BY %s " % (','.join(oNames)))

        if ((self.__limitStart is not None) and (self.__limitLength is not None)):
            oL.append(" LIMIT %d, %d " % (self.__limitStart, self.__limitLength))

        oL.append(";")
        #
        #
        return '\n'.join(oL)


class SqlGenCondition(object):

    """ Builds the Condition portion of an SQL selection or related query.
    """

    def __init__(self, schemaDefObj, addKeyJoinFlag=True, verbose=False):
        """  Input:

             schemaDef is instance of class derived from SchemaDefBase().
        """
        self.__schemaDefObj = schemaDefObj
        self.__verbose = verbose
        self.__ops = ['EQ', 'NE', 'GE', 'GT', 'LT', 'LE', 'LIKE', 'NOT LIKE', 'IS', 'IS NOT']
        self.__opDict = {'EQ': '=',
                         'NE': '!=',
                         'GE': '>=',
                         'GT': '>',
                         'LT': '<',
                         'LE': '<=',
                         'LIKE': 'LIKE',
                         'NOT LIKE': 'NOT LIKE',
                         'IS': 'IS',
                         'IS NOT': 'IS NOT'
                         }
        self.__logOps = ['AND', 'OR', 'NOT']
        self.__grpOps = ['BEGIN', 'END']
        #
        self.__cList = []
        self.__tableIdList = []
        self.__numConditions = 0
        self.__addKeyJoinFlag = addKeyJoinFlag
        #

    def clear(self):
        self.__cList = []
        self.__tableIdList = []
        self.__numConditions = 0
        return True

    def set(self, conditionDefList=None):
        """ Set/reset the current condition list --- The input is used verbatim and unmodified.
        """
        if conditionDefList is not None:
            self.__cList = conditionDefList
            self.__tableIdList = []
            for c in self.__cList:
                self.__updateTableList(c)
            return True
        else:
            return False

    def __updateTableList(self, cObj):
        """  Add the tables included in the input condition to the internal table list.
        """
        cType = cObj[0]
        if cType in ['VALUE_CONDITION', 'VALUE_LIST_CONDITION']:
            cType, lhsTuple, opCode, rhsTuple = cObj
            lTableId, lAttributeId = lhsTuple
            self.__addTable(lTableId)
            return True
        elif cType in ['JOIN_CONDITION']:
            cType, lhsTuple, opCode, rhsTuple = cObj
            lTableId, lAttributeId = lhsTuple
            rTableId, rAttributeId = rhsTuple
            self.__addTable(lTableId)
            self.__addTable(rTableId)
            return True
        else:
            return False

    def __addTable(self, tableId):
        if tableId not in self.__tableIdList:
            self.__tableIdList.append(tableId)
            return True
        else:
            return False

    def get(self):
        return self.__cList

    def getSql(self):
        if self.__addKeyJoinFlag:
            self.addKeyAttributeEquiJoinConditions()
        return self.__makeSql()

    def getTableIdList(self):
        return self.__tableIdList

    def addTables(self, tableIdList):
        '''  Add the tables from the input tableIdList to the internal list of tables.

             The internal list of tables is used to materialize join contraints between
             all tables based on primary keys defined in the schema definition.
        '''
        for tableId in tableIdList:
            self.__addTable(tableId)
        return True

    def addValueCondition(self, lhsTuple=None, opCode=None, rhsTuple=None, preOp='AND'):
        """  Adds a condition to the current contraint list -

             lhsTuple = (TableId,AttributeId)
             opCode   = one of the operations defined in self.__opDict.keys()
             rhsTuple = (value,type) where for

                        simple values -

                        (simpleValue,'CHAR|OTHER')
            preOp = logical operator preceding this contraint in the current contraint list.

        """
        cObj = ('VALUE_CONDITION', lhsTuple, opCode, rhsTuple)
        if cObj not in self.__cList:
            self.__updateTableList(cObj)
            if preOp in ['AND', 'OR']:
                self.addLogicalOp(lOp=preOp)
            self.addBeginGroup()
            self.__cList.append(cObj)
            self.addEndGroup()
            self.__numConditions += 1
        return self.__numConditions

    def addGroupValueConditionList(self, cDefList, preOp='AND'):
        """  Add a value alternative condition to the current contraint list
             using the input list of value condition definitions defined as -

             cDefList = [(lPreOp,lhsTuple, opCode, rhsTuple), ...]
               lPreOp   = local logical conjunction used to add condition within the group (leading value is ignored)
               lhsTuple = (TableId,AttributeId)
               opCode   = one of the operations defined in self.__opDict.keys()
               rhsTuple = (value,type) where for

                        simple values are defined as -

                        (simpleValue,'CHAR|<ANY OTHER>')   < CHAR > types are quoted

            preOp is the logical conjuction used to add the group condition to current condition list.

        """
        if len(cDefList) < 1:
            return self.__numConditions
        if preOp in ['AND', 'OR']:
            self.addLogicalOp(lOp=preOp)
        #
        self.addBeginGroup()
        for ii, cDef in enumerate(cDefList):
            (lPreOp, lhsTuple, opCode, rhsTuple) = cDef
            cObj = ('VALUE_CONDITION', lhsTuple, opCode, rhsTuple)
            self.__updateTableList(cObj)
            if ii > 0:
                self.addLogicalOp(lOp=lPreOp)
            self.addBeginGroup()
            self.__cList.append(cObj)
            self.addEndGroup()

            self.__numConditions += 1

        self.addEndGroup()
        return self.__numConditions

    def addJoinCondition(self, lhsTuple=None, opCode=None, rhsTuple=None, preOp='AND'):
        """  Adds a join condition to the current contraint list -

            lhsTuple = (TableId,AttributeId)
            opCode   = one of the operations defined in self.__opDict.keys()
            rhsTuple = (TableId,AttributeId)

                        For join conditions rhsTuple values implemented as -

                        (tableId,attributeId)

                        rhsTuple type is currrently only implemented as 'ATTRIBUTE' target but
                                 but could be extended to support other targets for different operators.

            preOp = logical operator preceding this contraint in the current contraint list.

        """
        cObj = ('JOIN_CONDITION', lhsTuple, opCode, rhsTuple)
        if cObj not in self.__cList:
            self.__updateTableList(cObj)
            if preOp in ['AND', 'OR']:
                self.addLogicalOp(lOp=preOp)
            self.addBeginGroup()
            self.__cList.append(cObj)
            self.addEndGroup()
            self.__numConditions += 1
        return self.__numConditions

    def addLogicalOp(self, lOp):
        """ Adds a logical operation into the current condition list.

             lOp  =  one of 'AND','OR', 'NOT'
        """
        self.__cList.append(('LOG_OP', lOp))

    def addBeginGroup(self):
        """ Inserts the beginning of a parenthetical group in the current condition list.
        """
        self.__cList.append(('GROUPING', 'BEGIN'))

    def addEndGroup(self):
        """ Inserts the ending of a parenthetical group in the current condition list.
        """
        self.__cList.append(('GROUPING', 'END'))

    def addKeyAttributeEquiJoinConditions(self):
        """ Auto add equi-join contraints between tables in the current table list -
        """
        self.__KeyJoinDone = True
        if len(self.__tableIdList) < 2:
            return 0
        cList = copy.deepcopy(self.__cList)
        self.__cList = []
        tablePairList = [t for t in itertools.combinations(self.__tableIdList, 2)]
        for (t1, t2) in tablePairList:
            self.__addInterTableJoinContraints(t1, t2)
        for c in cList:
            if c[0] in ['JOIN_CONDITION'] and c in self.__cList:
                continue
            self.__cList.append(c)

        return len(tablePairList)

    def __makeSql(self):
        """  Builds SQL string for the query condition encoded in the input contraint command list.

             The condition command list is a sequence of tuples with the following syntax:

             ('VALUE_CONDITION', (tableId,attributeId), 'EQ'|'NE'|GE'|'GT'|'LT'|'LE'|'LIKE'|'IS'|'IS NOT', (Value, 'CHAR'|'OTHER'))

             ('VALUE_LIST_CONDITION', (tableId,attributeId), 'IN'|'NOT IN', (valueList, 'CHAR'|'INT'|'FLOAT'))

             ('JOIN_CONDITION',  (tableId,attributeId), 'EQ'|'NE'|'GE'|'GT'|'LT'|'LE'|'LIKE', (tableId,attributeId) )

             ('LOG_OP', 'AND'|'OR'|'NOT')     -> conjunction / negation

             ('GROUPING', 'BEGIN'|'END')          -> (grouping/parenthetical control)

        """
        #
        cSqlL = []
        #
        #
        cCount = 0
        for c in self.__cList:
            cType = (c[0]).upper()
            if cType == 'VALUE_CONDITION':
                cCount += 1
                (cT, lhsTuple, opId, rhsTuple) = c
                (tableId, attributeId) = lhsTuple
                (value, vType) = rhsTuple
                vType = vType.upper()
                qAttributeName = self.__schemaDefObj.getQualifiedAttributeName(tableAttributeTuple=lhsTuple)
                # jdw quote date and datetime
                if (vType in ['CHAR', 'VARCHAR', 'DATE', 'DATETIME']):
                    cSqlL.append(" %s %s '%s' " % (qAttributeName, self.__opDict[opId], value))
                else:
                    cSqlL.append(" %s %s %s " % (qAttributeName, self.__opDict[opId], value))

            elif cType == 'VALUE_LIST_CONDITION':
                cCount += 1
                (cT, lhsTuple, opId, rhsTuple) = c
                (tableId, attributeId) = lhsTuple
                (valueList, vType) = rhsTuple
                vType = vType.upper()
                qAttributeName = self.__schemaDefObj.getQualifiedAttributeName(tableAttributeTuple=lhsTuple)
                #
                qVL = []
                for value in valueList:
                    if vType in ['CHAR', 'VARCHAR']:
                        qV = "'%s'" % value
                    elif vType in ['INT', 'INTEGER']:
                        qV = "%i" % value
                    elif vType in ['FLOAT', 'DOUBLE', 'DECIMAL']:
                        qV = "%f" % value
                    else:
                        qV = value
                    qVL.append(qV)
                vS = ','.join
                cSqlL.append(" %s %s [ %s ] " % (qAttributeName, self.__opDict[opId], vS))

            elif cType == 'JOIN_CONDITION':
                cCount += 1
                (cT, lhsTuple, opId, rhsTuple) = c
                (lTableId, lAttributeId) = lhsTuple
                (rTableId, rAttributeId) = rhsTuple
                lAttributeName = self.__schemaDefObj.getQualifiedAttributeName(tableAttributeTuple=lhsTuple)
                rAttributeName = self.__schemaDefObj.getQualifiedAttributeName(tableAttributeTuple=rhsTuple)
                #
                cSqlL.append(" %s %s %s " % (lAttributeName, self.__opDict[opId], rAttributeName))

            elif cType == 'LOG_OP':
                (cT, logOp) = c
                if logOp not in ['NOT'] and cCount > 0:
                    cSqlL.append(" %s " % logOp)
            elif cType == 'GROUPING':
                (cT, group) = c
                if group == 'BEGIN':
                    cSqlL.append(" (  ")
                elif group == 'END':
                    cSqlL.append(" ) ")
            else:
                pass

        return '\n'.join(cSqlL)

    def __addInterTableJoinContraints(self, lTableId, rTableId):
        """ The ...
        """
        lTdef = self.__schemaDefObj.getSchemaObject(lTableId)
        lKeyAttributeIdL = lTdef.getPrimaryKeyAttributeIdList()
        rTdef = self.__schemaDefObj.getSchemaObject(rTableId)
        rKeyAttributeIdL = rTdef.getPrimaryKeyAttributeIdList()
        #
        commonAttributeIdSet = set(lKeyAttributeIdL) & set(rKeyAttributeIdL)
        logger.debug("+SqlGenCondition.__addInterTableJoinConditions lTable %s rTable %s  common keys %r\n" % (lTableId, rTableId, commonAttributeIdSet))
        #
        for attributeId in commonAttributeIdSet:
            lhsTuple = (lTableId, attributeId)
            opCode = 'EQ'
            rhsTuple = (rTableId, attributeId)
            self.addJoinCondition(lhsTuple=lhsTuple, opCode=opCode, rhsTuple=rhsTuple, preOp='AND')
        #
        return len(commonAttributeIdSet)



##
# File:    RepoPathUtil.py
# Author:  J. Westbrook
# Date:    21-Mar-2018
# Version: 0.001
#
# Updates:
#   22-Mar-2018  jdw add support for all repositories -
#   26-Mar-2018  jdw internalize the use of externally provided configuration object -
#   27-Mar-2018  jdw add path to support mock repositories for testing.
#   23-May-2018  jdw add getRepoPathList() convenience method
#   18-Jun-2018  jdw move mock support to the configuration module
#   12-Jul-2018  jdw correct config for PDBX_REPO_PATH
#   13-Aug-2018  jdw add support for gz compressed entry files
#   24-Oct-2018  jdw update for new configuration organization
##
"""
 Utilites for scanning common data repository file systems.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import time

from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

try:
    from os import walk
except ImportError:
    from scandir import walk


logger = logging.getLogger(__name__)


class RepoPathUtil(object):

    def __init__(self, cfgOb, cfgSectionName='site_info', numProc=8, fileLimit=None, verbose=False):
        self.__fileLimit = fileLimit
        self.__numProc = numProc
        self.__verbose = verbose
        self.__cfgOb = cfgOb
        self.__cfgSectionName = cfgSectionName
        self.__mpFormat = '[%(levelname)s] %(asctime)s %(processName)s-%(module)s.%(funcName)s: %(message)s'

    def getRepoPathList(self, contentType, inputPathList=None):
        """ Convenience method to return repository path list by content type:
        """
        outputPathList = []
        inputPathList = inputPathList if inputPathList else []
        try:
            if contentType == "bird":
                outputPathList = inputPathList if inputPathList else self.getBirdPathList()
            elif contentType == "bird_family":
                outputPathList = inputPathList if inputPathList else self.getBirdFamilyPathList()
            elif contentType in ['chem_comp', 'chem_comp_core']:
                outputPathList = inputPathList if inputPathList else self.getChemCompPathList()
            elif contentType in ['bird_chem_comp', 'bird_chem_comp_core']:
                outputPathList = inputPathList if inputPathList else self.getBirdChemCompPathList()
            elif contentType in ['pdbx', 'pdbx_core']:
                outputPathList = inputPathList if inputPathList else self.getEntryPathList()
            elif contentType in ['pdb_distro', 'da_internal', 'status_history']:
                outputPathList = inputPathList if inputPathList else []
            else:
                logger.warning("Unsupported contentType %s" % contentType)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        if self.__fileLimit:
            outputPathList = outputPathList[:self.__fileLimit]

        return outputPathList

    def _chemCompPathWorker(self, dataList, procName, optionsD, workingDir):
        """ Return the list of chemical component definition file paths in the current repository.
        """
        topRepoPath = optionsD['topRepoPath']
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, dirs, files in walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.endswith(".cif") and len(name) <= 7:
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def getChemCompPathList(self):
        return self.__getChemCompPathList(self.__cfgOb.getPath('CHEM_COMP_REPO_PATH', sectionName=self.__cfgSectionName), numProc=self.__numProc)

    def __getChemCompPathList(self, topRepoPath, numProc=8):
        """Get the path list for the chemical component definition repository
        """
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s" % ts)
        startTime = time.time()
        pathList = []
        try:
            dataS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'
            dataList = [a for a in dataS]
            optD = {}
            optD['topRepoPath'] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_chemCompPathWorker")
            ok, failList, retLists, diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds\n" % (len(pathList), endTime0 - startTime))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return self.__applyFileLimit(pathList)

    def _entryPathWorker(self, dataList, procName, optionsD, workingDir):
        """ Return the list of entry file paths in the current repository.
        """
        topRepoPath = optionsD['topRepoPath']
        pathList = []
        for subdir in dataList:
            dd = os.path.join(topRepoPath, subdir)
            for root, dirs, files in walk(dd, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if ((name.endswith(".cif.gz") and len(name) == 11) or (name.endswith(".cif") and len(name) == 8)):
                        pathList.append(os.path.join(root, name))
        return dataList, pathList, []

    def getEntryPathList(self):
        return self.__getEntryPathList(self.__cfgOb.getPath('PDBX_REPO_PATH', sectionName=self.__cfgSectionName), numProc=self.__numProc)

    def __getEntryPathList(self, topRepoPath, numProc=8):
        """Get the path list for structure entries in the input repository
        """
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting at %s" % ts)
        startTime = time.time()
        pathList = []
        try:
            dataList = []
            anL = 'abcdefghijklmnopqrstuvwxyz0123456789'
            for a1 in anL:
                for a2 in anL:
                    hc = a1 + a2
                    dataList.append(hc)
                    hc = a2 + a1
                    dataList.append(hc)
            dataList = list(set(dataList))
            #
            optD = {}
            optD['topRepoPath'] = topRepoPath
            mpu = MultiProcUtil(verbose=self.__verbose)
            mpu.setOptions(optionsD=optD)
            mpu.set(workerObj=self, workerMethod="_entryPathWorker")
            ok, failList, retLists, diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            pathList = retLists[0]
            endTime0 = time.time()
            logger.debug("Path list length %d  in %.4f seconds\n" % (len(pathList), endTime0 - startTime))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return self.__applyFileLimit(pathList)

    def getBirdPathList(self):
        return self.__getBirdPathList(self.__cfgOb.getPath('BIRD_REPO_PATH', sectionName=self.__cfgSectionName))

    def __getBirdPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRD_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def getBirdFamilyPathList(self):
        return self.__getBirdFamilyPathList(self.__cfgOb.getPath('BIRD_FAMILY_REPO_PATH', sectionName=self.__cfgSectionName))

    def __getBirdFamilyPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("FAM_") and name.endswith(".cif") and len(name) <= 14:
                        pth = os.path.join(root, name)
                        sd[int(name[4:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def getBirdChemCompPathList(self):
        return self.__getBirdChemCompPathList(self.__cfgOb.getPath('BIRD_CHEM_COMP_REPO_PATH', sectionName=self.__cfgSectionName))

    def __getBirdChemCompPathList(self, topRepoPath):
        """ Return the list of definition file paths in the current repository.

            List is ordered in increasing PRD ID numerical code.
        """
        pathList = []
        try:
            sd = {}
            for root, dirs, files in os.walk(topRepoPath, topdown=False):
                if "REMOVE" in root:
                    continue
                for name in files:
                    if name.startswith("PRDCC_") and name.endswith(".cif") and len(name) <= 16:
                        pth = os.path.join(root, name)
                        sd[int(name[6:-4])] = pth
            #
            for k in sorted(sd.keys()):
                pathList.append(sd[k])
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        #
        return self.__applyFileLimit(pathList)

    def __applyFileLimit(self, pathList):
        logger.debug("Length of file path list %d (limit %r)" % (len(pathList), self.__fileLimit))
        if self.__fileLimit:
            return pathList[:self.__fileLimit]
        else:
            return pathList



##
# File: MysqlSchemaImporter.py
# Date: 20-May-2015 jdw
#
# Create a skeleton schema map definition data structure from mysql
# table descriptions.
#
# ** This is a command-line utility that is not part of the overall api
#
import logging
import os
import pprint
import sys

logger = logging.getLogger(__name__)


class MysqlSchemaImporter(object):

    def __init__(self, dbUser, dbPw, dbHost, mysqlPath='/opt/local/bin/mysql', verbose=True):
        self.__verbose = verbose
        self.__mysqlPath = mysqlPath
        self.__dbUser = dbUser
        self.__dbPw = dbPw
        self.__dbHost = dbHost

    def __import(self, filePath):
        colDataList = []
        ifh = open(filePath, 'r')
        for line in ifh:
            if line is not None and len(line) > 0:
                fields = str(line[:-1]).split('\t')
                if len(fields) != 6:
                    logger.info("bad line in %s  = %s" % (filePath, line))
                    continue
                else:
                    colDataList.append(fields)
        ifh.close()
        os.remove(filePath)
        #
        return colDataList[1:]

    def __export(self, filePath, db, tableName):
        cmdDetail = ' --user=%s --password=%s --host=%s %s -e "describe %s;" ' % (self.__dbUser, self.__dbPw, self.__dbHost, db, tableName)
        cmd = self.__mysqlPath + cmdDetail + ' > %s' % filePath
        return os.system(cmd)

    def __buildDef(self, dbName, tableName, colDataList):
        defD = {}
        tableId = str(tableName).upper()
        attIdKeyList = []
        attMap = {}
        indD = {}
        attInfo = {}
        attD = {}
        for ii, ff in enumerate(colDataList, start=1):
            attName = str(ff[0])
            attId = str(attName).upper()
            nullFlag = True if ff[2] == 'YES' else False
            impType = ff[1]
            if '(' in impType:
                width = impType[impType.find('(') + 1:-1]
                sqlType = impType[:impType.find('(')]
            else:
                width = 10
                sqlType = impType
            precision = 0
            if ff[3] in ['MUL', 'PRI']:
                attIdKeyList.append(attId)
                keyFlag = True
            else:
                keyFlag = False
            attD[attId] = attName
            attMap[attId] = (tableName, attName, None, None)
            attInfo[attId] = {'NULLABLE': nullFlag,
                              'ORDER': ii,
                              'PRECISION': precision,
                              'PRIMARY_KEY': keyFlag,
                              'APP_TYPE': sqlType.upper(),
                              'WIDTH': width}
        #
        d = {}
        d['ATTRIBUTES'] = tuple(attIdKeyList)
        d['TYPE'] = 'UNIQUE'
        indD['p1'] = d
        defD['INDICES'] = indD
        defD['ATTRIBUTES'] = attD
        defD['ATTRIBUTE_INFO'] = attInfo
        defD['ATTRIBUTE_MAP'] = attMap
        #
        defD['SCHEMA_DELETE_ATTRIBUTE'] = attIdKeyList[0]
        defD['SCHEMA_ID'] = tableId
        defD['SCHEMA_NAME'] = tableName
        defD['SCHEMA_TYPE'] = 'transactional'
        # 'MAP_MERGE_INDICES': {'valence_ref': {'ATTRIBUTES': ('id',), 'TYPE': 'EQUI-JOIN'}},
        tD = {}
        tD['ATTRIBUTES'] = tuple(attIdKeyList)
        tD['TYPE'] = 'EQUI-JOIN'
        defD['MAP_MERGE_INDICES'] = {}
        defD['MAP_MERGE_INDICES'][tableName] = tD
        #
        return tableId, defD

    def create(self, dbName, tableNameList):
        schemaDef = {}
        for tableName in tableNameList:
            fn = 'mysql-schema-' + tableName + '.txt'
            self.__export(fn, dbName, tableName)
            colDataList = self.__import(fn)
            if colDataList and len(colDataList) > 0:
                logger.info("tableName %s length %d\n" % (tableName, len(colDataList)))
                tableId, defD = self.__buildDef(dbName, tableName, colDataList)
                schemaDef[tableId] = defD
        #
        pprint.pprint(schemaDef, stream=sys.stdout, width=120, indent=3)


def importxxx(self):
    tableNameList = ["entity",
                     "entity_poly",
                     "entity_src_gen",
                     "entity_src_nat",
                     "pdbx_entity_src_syn"]
    # tableNameList = ['weight_in_asu']
    dbUser = os.getenv('MYSQL_DB_USER')
    dbPw = os.getenv('MYSQL_SBKB_PW')
    dbHost = 'localhost'
    msi = MysqlSchemaImporter(dbUser, dbPw, dbHost, mysqlPath='/opt/local/bin/mysql', verbose=True)
    msi.create('stat', tableNameList)


if __name__ == "__main__":

    tableNameList = ['PDB_status_information',
                     'audit_author',
                     'chem_comp',
                     'citation',
                     'citation_author',
                     'deposition_from_09',
                     'deposition_from_2',
                     'entity',
                     'entity_poly',
                     'pdb_entry',
                     'pdbx_contact_author',
                     'pdbx_database_PDB_obs_spr',
                     'pdbx_database_status_history',
                     'pdbx_depui_entry_details',
                     'pdbx_molecule',
                     'pdbx_molecule_features',
                     'pdbx_prerelease_seq',
                     'processing_status',
                     'rcsb_status',
                     'rcsb_status_t',
                     'struct',
                     'pdbx_database_status_history']

    dbUser = os.getenv('MYSQL_DB_USER')
    dbPw = os.getenv('MYSQL_SBKB_PW')
    dbHost = 'localhost'
    msi = MysqlSchemaImporter(dbUser, dbPw, dbHost, mysqlPath='/opt/local/bin/mysql', verbose=True)
    msi.create('da_internal', tableNameList)



##
# File:    ScanRepoUtil.py
# Author:  J. Westbrook
# Date:    30-Apr-2018
# Version: 0.001
#
# Updates:
#  9-May-2018 jdw implement incremental update of scanned data
# 18-May-2018 jdw move IO to IoUtils.py
# 23-May-2018 jdw remove proxy to ContentTypeUtil(), go directly to RepoPathUtil() for paths
#  5-Jun-2018 jdw update prototypes for IoUtil() methods
# 16-Jun-2018 jdw update data type prototype.
# 18-Jun-2018 jdw move mocking to configuration level
# 28-Jun-2018 jdw remove IoUtil() and add working path constructor argument
##
"""
Tools for for scanning repositories and collecting coverage and type data information.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import collections
import datetime
import logging
import time

from mmcif.api.DataCategory import DataCategory

from rcsb.db.utils.RepoPathUtil import RepoPathUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)

ScanValue = collections.namedtuple('ScanValue', 'containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec')
ScanSummary = collections.namedtuple('ScanSummary', 'containerId, fromPath, scanDate, scanCategoryDict')


class ScanRepoUtil(object):
    """Tools for for scanning repositories and collecting coverage and type data information.

    """

    def __init__(self, cfgOb, attributeDataTypeD=None, numProc=4, chunkSize=15, fileLimit=None, maxStepLength=2000, workPath=None):
        """
        Args:
            cfgOb (object): Configuration object (ConfigUtil)

            attributeDataTypeD
            dictPath (str): Path to supporting data dictionary

            numProc (int, optional): Number of parallel worker processes used.
            chunkSize (int, optional): Size of files processed in a single multi-proc process
            fileLimit (int, optional): maximum file scanned or None for no limit
            mockTopPath (str, optional): Path to directory containing mock repositories or None
            maxStepLength (int, optional): maximum number of multi-proc runs to perform
        """
        #
        self.__attributeDataTypeD = attributeDataTypeD if attributeDataTypeD else {}
        # Limit the load length of each file type for testing  -  Set to None to remove -
        self.__fileLimit = fileLimit
        self.__maxStepLength = maxStepLength
        #
        # Controls for multiprocessing execution -
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        #
        self.__cfgOb = cfgOb
        #
        self.__mpFormat = '[%(levelname)s] %(asctime)s %(processName)s-%(module)s.%(funcName)s: %(message)s'

        self.__workPath = workPath
        self.__mU = MarshalUtil(workPath=self.__workPath)

    def scanContentType(self, contentType, scanType='full', inputPathList=None, scanDataFilePath=None, failedFilePath=None, saveInputFileListPath=None):
        """Driver method for scan operation

        Args:
            contentType (str):  one of 'bird','bird_family','bird_chem_comp', chem_comp','pdbx'
            scanType (str, optional): 'full' [or 'incr' to be supported]
            inputPathList (list, optional):  list of input file paths to scan
            scanDataFilePath (str, optional): file path for serialized scan data (Pickle format)
            failedFilePath (str, optional): file path for list of files that fail scanning operation
            saveInputFileListPath str, optional): Path to store file path list that is scanned

        Returns:
            bool: True for success or False otherwise

        """
        try:
            startTime = self.__begin(message="scanning operation")
            #
            rpU = RepoPathUtil(self.__cfgOb, numProc=self.__numProc, fileLimit=self.__fileLimit)
            pathList = rpU.getRepoPathList(contentType=contentType, inputPathList=inputPathList)
            #

            if saveInputFileListPath:
                self.__mU.doExport(saveInputFileListPath, pathList, format="list")
                logger.debug("Saving %d paths in %s" % (len(pathList), saveInputFileListPath))
            #

            optD = {}
            optD['contentType'] = contentType
            optD['logSize'] = True
            optD['scanType'] = scanType
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            #

            numProc = self.__numProc
            chunkSize = self.__chunkSize if inputPathList and self.__chunkSize < len(inputPathList) else 0
            #
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            numPaths = len(pathList)
            logger.debug("Processing %d total paths" % numPaths)
            numProc = min(numProc, numPaths)
            maxStepLength = self.__maxStepLength
            if numPaths > maxStepLength:
                numLists = int(numPaths / maxStepLength)
                subLists = [pathList[i::numLists] for i in range(numLists)]
            else:
                subLists = [pathList]
            #
            if subLists and len(subLists) > 0:
                logger.debug("Starting with numProc %d outer subtask count %d subtask length ~ %d" % (numProc, len(subLists), len(subLists[0])))
            #
            numResults = 1
            failList = []
            retLists = [[] for ii in range(numResults)]
            diagList = []
            for ii, subList in enumerate(subLists):
                logger.debug("Running outer subtask %d or %d length %d" % (ii + 1, len(subLists), len(subList)))
                #
                mpu = MultiProcUtil(verbose=True)
                mpu.setOptions(optionsD=optD)
                mpu.set(workerObj=self, workerMethod="scanWorker")
                ok, failListT, retListsT, diagListT = mpu.runMulti(dataList=subList, numProc=numProc, numResults=numResults, chunkSize=chunkSize)
                failList.extend(failListT)
                # retLists is a list of lists -
                for ii in range(numResults):
                    retLists[ii].extend(retListsT[ii])
                diagList.extend(diagListT)
            logger.debug("Scan failed path list %r" % failList)
            logger.debug("Scan path list success length %d load list failed length %d" % (len(pathList), len(failList)))
            logger.debug("Returned metadata length %r" % len(retLists[0]))
            #
            if failedFilePath and len(failList):
                wOk = self.__mU.doExport(failedFilePath, failList, format="list")
                logger.debug("Writing scan failure path list to %s status %r" % (failedFilePath, wOk))
            #
            if scanType == 'incr':
                scanDataD = self.__mU.doImport(scanDataFilePath, format='pickle', default=None)
                logger.info("Imported scan data with keys %r" % list(scanDataD.keys()))
            else:
                scanDataD = {}
            #
            if scanDataFilePath and len(retLists[0]):
                for ssTup in retLists[0]:
                    cId = ssTup.containerId
                    if scanType == 'full' and cId in scanDataD:
                        logger.error("Duplicate container id %s in %r and %r" % (cId, ssTup.fromPath, scanDataD[cId].fromPath))
                    #
                    scanDataD[cId] = ssTup

                ok = self.__mU.doExport(scanDataFilePath, scanDataD, format='pickle')
                tscanDataD = self.__mU.doImport(scanDataFilePath, format='pickle')
                ok = tscanDataD == scanDataD

            self.__end(startTime, "scanning operation with status " + str(ok))

            #
            return ok
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return False

    def evalScan(self, scanDataFilePath, evalJsonFilePath, evalType='data_type'):

        scanDataD = self.__mU.doImport(scanDataFilePath, format='pickle')
        if evalType in ['data_type']:
            rD = self.__evalScanDataType(scanDataD)
        elif evalType in ['data_coverage']:
            rD = self.__evalScanDataCoverage(scanDataD)
        else:
            logger.debug("Unknown evalType %r " % evalType)
        ok = self.__mU.doExport(evalJsonFilePath, rD, format='json')

        return ok

    def __evalScanDataType(self, scanDataD):
        """
        ScanValue = collections.namedtuple('ScanValue', 'containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec')
        ScanSummary = collections.namedtuple('ScanSummary', 'containerId, fromPath, scanDate, scanCategoryDict')

        """
        # for populated sD[category] -> d[atName]->{minWidth: , maxWidth:, minPrec:, maxPrec: , count}
        sD = {}
        for cId in scanDataD:
            ssTup = scanDataD[cId]
            d = ssTup.scanCategoryDict
            for catName in d:
                if catName not in sD:
                    sD[catName] = {}
                for svTup in d[catName]:
                    if svTup.atName not in sD[catName]:
                        sD[catName][svTup.atName] = {'minWidth': svTup.minWidth, 'maxWidth': svTup.maxWidth, 'minPrec': svTup.minPrec, 'maxPrec': svTup.maxPrec, 'count': 1}
                        continue
                    sD[catName][svTup.atName]['minWidth'] = min(sD[catName][svTup.atName]['minWidth'], svTup.minWidth)
                    sD[catName][svTup.atName]['maxWidth'] = max(sD[catName][svTup.atName]['maxWidth'], svTup.maxWidth)
                    sD[catName][svTup.atName]['minPrec'] = min(sD[catName][svTup.atName]['minPrec'], svTup.minPrec)
                    sD[catName][svTup.atName]['maxPrec'] = max(sD[catName][svTup.atName]['maxPrec'], svTup.maxPrec)
                    sD[catName][svTup.atName]['count'] += 1
        return sD

    def __evalScanDataCoverage(self, scanDataD):
        """
        ScanValue = collections.namedtuple('ScanValue', 'containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec')
        ScanSummary = collections.namedtuple('ScanSummary', 'containerId, fromPath, scanDate, scanCategoryDict')

        """

        # for populated sD[category] -> d[atName]->{count: #, instances: [id,id,id]}
        sD = {}
        for cId in scanDataD:
            ssTup = scanDataD[cId]
            d = ssTup.scanCategoryDict
            for catName in d:
                if catName not in sD:
                    sD[catName] = {}
                for svTup in d[catName]:
                    if svTup.atName not in sD[catName]:
                        sD[catName][svTup.atName] = {'count': 0, 'instances': []}
                    sD[catName][svTup.atName]['instances'].append(svTup.containerId)
                    sD[catName][svTup.atName]['count'] += 1
        return sD

    def scanWorker(self, dataList, procName, optionsD, workingDir):
        """ Multi-proc worker method for scanning repository data files-


        """
        try:
            startTime = self.__begin(message=procName)
            # Recover common options

            scanType = optionsD['scanType']
            contentType = optionsD['contentType']
            #
            successList = []
            retList = []

            containerList = self.__getContainerList(dataList)
            for container in containerList:
                ret = self.__scanContainer(container)
                successList.append(ret.fromPath)
                retList.append(ret)
            #

            logger.debug("%s scanType %s contentType %spathlist length %d containerList length %d" % (procName, scanType, contentType, len(dataList), len(containerList)))

            ok = len(successList) == len(dataList)
            #
            self.__end(startTime, procName + " with status " + str(ok))
            return successList, retList, []

        except Exception as e:
            logger.error("Failing with dataList %r" % dataList)
            logger.exception("Failing with %s" % str(e))

        return [], [], []

    def __getContainerList(self, inputPathList):
        """
        """
        utcnow = datetime.datetime.utcnow()
        ts = utcnow.strftime("%Y-%m-%d:%H:%M:%S")

        cL = []
        for lPath in inputPathList:
            myContainerList = self.__mU.doImport(lPath, format="mmcif")

            for c in myContainerList:
                dc = DataCategory('rcsb_load_status', ['name', 'load_date', 'locator'], [[c.getName(), ts, lPath]])
                logger.debug("data category %r" % dc)
                c.append(dc)
                cL.append(c)
        return cL

    def __scanContainer(self, container):
        """ Scan the input container for

          Get the file name -
        """
        cName = container.getName()
        loadStatusObj = container.getObj('rcsb_load_status')
        lName = loadStatusObj.getValue(attributeName='name', rowIndex=0)
        lFilePath = loadStatusObj.getValue(attributeName='locator', rowIndex=0)
        lDate = loadStatusObj.getValue(attributeName='load_date', rowIndex=0)
        #
        oD = {}
        for objName in container.getObjNameList():
            if objName == 'rcsb_load_status':
                continue
            obj = container.getObj(objName)
            afD = self.__attributeDataTypeD[objName] if objName in self.__attributeDataTypeD else {}
            atNameList = obj.getAttributeList()
            wMin = {atName: 100000 for atName in atNameList}
            wMax = {atName: -1 for atName in atNameList}
            pMin = {atName: 100000 for atName in atNameList}
            pMax = {atName: -1 for atName in atNameList}
            for row in obj.getRowList():
                for ii, val in enumerate(row):
                    valLen = len(val)
                    if ((valLen == 0) or (val == '?') or (val == '.')):
                        continue
                    atName = atNameList[ii]
                    wMin[atName] = min(wMin[atName], valLen)
                    wMax[atName] = max(wMax[atName], valLen)
                    if atName in afD and afD[atName] == 'float':
                        vPrec = 0
                        try:
                            fields = val.split('.')
                            vPrec = len(fields[1])
                            pMin[atName] = min(pMin[atName], vPrec)
                            pMax[atName] = max(pMax[atName], vPrec)
                        except Exception as e:
                            logger.debug("Failed to process float %s %r %r %s" % (atName, val, vPrec, str(e)))
                            pMin[atName] = 0
                            pMax[atName] = 0
                        logger.debug("Got float for %s %r %r" % (atName, val, vPrec))
                    else:
                        pMin[atName] = 0
                        pMax[atName] = 0

            # ScanValue - containerId, catName, atName, minWidth, maxWidth, minPrec, maxPrec
            oD[objName] = [ScanValue(cName, objName, atN, wMin[atN], wMax[atN], pMin[atN], pMax[atN]) for atN in wMax if wMax[atN] != -1]
        # ScanSummary containerId, fromPath, scanCategoryDict
        #
        ret = ScanSummary(lName, lFilePath, lDate, oD)
        #
        return ret

    def __begin(self, message=""):
        startTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting %s at %s" % (message, ts))
        return startTime

    def __end(self, startTime, message=""):
        endTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        delta = endTime - startTime
        logger.debug("Completed %s at %s (%.4f seconds)" % (message, ts, delta))



##
# File: RepoLoadExec.py
# Date: 15-Mar-2018  jdw
#
#  Execution wrapper  --  repository database loading utilities --
#  Updates:
#
#    21-Mar-2018 - jdw added content filters and separate collection for Bird chemical components
#    22-May-2018 - jdw add replacment load type, add options for input file paths
#    24-Mar-2018 - jdw add split collections for entries (preiliminary)
#    24-Mar-2018 - jdw add option to preserve paths from automatic repo scan
#    27-Mar-2018 - jdw update configuration handling and add support for mocking repository paths
#     4-Apr-2018 - jdw added option to prune documents to a size limit
#     3-Jul-2018 - jdw specialize for repository loading.
#    14-Jul-2018 - jdw add loading of separate
#    25-Jul-2018 - jdw die on input file path list processing error
#    20-Aug-2018 - jdw add load_pdbx_core load option
#     9-Sep-2018 - jdw expose --schema_level option
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.mongo.PdbxLoader import PdbxLoader
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()


def loadStatus(statusList, cfgOb, readBackCheck=True):
    sectionName = 'data_exchange'
    dl = DocumentLoader(cfgOb, "MONGO_DB", numProc=2, chunkSize=2,
                        documentLimit=None, verbose=False, readBackCheck=readBackCheck)
    #
    databaseName = cfgOb.get('DATABASE_NAME', sectionName=sectionName) + '_' + cfgOb.get('DATABASE_VERSION_STRING', sectionName=sectionName)
    collectionVersion = cfgOb.get('COLLECTION_VERSION_STRING', sectionName=sectionName)
    collectionName = cfgOb.get('COLLECTION_UPDATE_STATUS', sectionName=sectionName) + '_' + collectionVersion
    ok = dl.load(databaseName, collectionName, loadType='append', documentList=statusList,
                 indexAttributeList=['update_id', 'database_name', 'object_name'], keyNames=None)
    return ok


def main():
    parser = argparse.ArgumentParser()
    #
    #
    parser.add_argument("--full", default=False, action='store_true', help="Fresh full load in a new tables/collections")
    parser.add_argument("--replace", default=False, action='store_true', help="Load with replacement in an existing table/collection (default)")
    #
    parser.add_argument("--load_chem_comp_ref", default=False, action='store_true', help="Load Chemical Component reference definitions (public subset)")
    parser.add_argument("--load_chem_comp_core_ref", default=False, action='store_true', help="Load Chemical Component Core reference definitions (public subset)")
    parser.add_argument("--load_bird_chem_comp_ref", default=False, action='store_true', help="Load Bird Chemical Component reference definitions (public subset)")
    parser.add_argument("--load_bird_chem_comp_core_ref", default=False, action='store_true', help="Load Bird Chemical Component Core reference definitions (public subset)")
    parser.add_argument("--load_bird_ref", default=False, action='store_true', help="Load Bird reference definitions (public subset)")
    parser.add_argument("--load_bird_family_ref", default=False, action='store_true', help="Load Bird Family reference definitions (public subset)")
    parser.add_argument("--load_entry_data", default=False, action='store_true', help="Load PDBx entry data (current released subset)")
    parser.add_argument("--load_pdbx_core", default=False, action='store_true', help="Load PDBx core entry/entity data (current released subset)")
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default="site_info", help="Configuration section name")

    parser.add_argument("--db_type", default="mongo", help="Database server type (default=mongo)")

    parser.add_argument("--document_style", default="rowwise_by_name_with_cardinality",
                        help="Document organization (rowwise_by_name_with_cardinality|rowwise_by_name|columnwise_by_name|rowwise_by_id|rowwise_no_name")
    parser.add_argument("--read_back_check", default=False, action='store_true', help="Perform read back check on all documents")
    parser.add_argument("--schema_level", default=None, help="Schema validation level (full|min default=None)")
    #
    parser.add_argument("--load_file_list_path", default=None, help="Input file containing load file path list (override automatic repository scan)")
    parser.add_argument("--fail_file_list_path", default=None, help="Output file containing file paths that fail to load")
    parser.add_argument("--save_file_list_path", default=None, help="Save repo file paths from automatic file system scan in this path")

    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--file_limit", default=None, help="Load file limit for testing")
    parser.add_argument("--prune_document_size", default=None, help="Prune large documents to this size limit (MB)")
    parser.add_argument("--debug", default=False, action='store_true', help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action='store_true', help="Use MOCK repository configuration for testing")
    parser.add_argument("--working_path", default=None, help="Working path for temporary files")
    args = parser.parse_args()
    #
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #                                       Configuration Details
    configPath = args.config_path
    configName = args.config_name
    if not configPath:
        configPath = os.getenv('DBLOAD_CONFIG_PATH', None)
    try:
        if os.access(configPath, os.R_OK):
            os.environ['DBLOAD_CONFIG_PATH'] = configPath
            logger.info("Using configuation path %s (%s)" % (configPath, configName))
        else:
            logger.error("Missing or access issue with config file %r" % configPath)
            exit(1)
        mockTopPath = os.path.join(TOPDIR, 'rcsb', 'mock-data') if args.mock else None
        cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=mockTopPath)
    except Exception as e:
        logger.error("Missing or access issue with config file %r with %s" % (configPath, str(e)))
        exit(1)

    #
    try:
        readBackCheck = args.read_back_check
        numProc = int(args.num_proc)
        chunkSize = int(args.chunk_size)
        fileLimit = int(args.file_limit) if args.file_limit else None
        failedFilePath = args.fail_file_list_path
        fPath = args.load_file_list_path
        schemaLevel = args.schema_level if args.schema_level in ['min', 'full', 'minimum'] else None
        loadType = 'full' if args.full else 'replace'
        loadType = 'replace' if args.replace else 'full'
        saveInputFileListPath = args.save_file_list_path
        pruneDocumentSize = float(args.prune_document_size) if args.prune_document_size else None
        workPath = args.working_path if args.working_path else '.'
        if args.document_style not in ['rowwise_by_name', 'rowwise_by_name_with_cardinality', 'columnwise_by_name', 'rowwise_by_id', 'rowwise_no_name']:
            logger.error("Unsupported document style %s" % args.document_style)
        if args.db_type != "mongo":
            logger.error("Unsupported database server type %s" % args.db_type)
    except Exception as e:
        logger.exception("Argument processing problem %s" % str(e))
        parser.print_help(sys.stderr)
        exit(1)
    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -
    #
    # Read any input path lists -
    #

    inputPathList = None
    if fPath:
        mu = MarshalUtil(workPath=workPath)
        inputPathList = mu.doImport(fPath, format='list')
        if len(inputPathList) < 1:
            logger.error("Missing or empty input file path list %s" % fPath)
            exit(1)
    #
    ##
    if args.db_type == "mongo":
        mw = PdbxLoader(
            cfgOb,
            cfgSectionName=configName,
            resourceName="MONGO_DB",
            numProc=numProc,
            chunkSize=chunkSize,
            fileLimit=fileLimit,
            verbose=debugFlag,
            readBackCheck=readBackCheck,
            workPath=workPath)

        if args.load_chem_comp_ref:
            ok = mw.load('chem_comp', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_chem_comp_core_ref:
            ok = mw.load('chem_comp_core', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_bird_chem_comp_ref:
            ok = mw.load('bird_chem_comp', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_bird_chem_comp_core_ref:
            ok = mw.load('bird_chem_comp_core', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_bird_ref:
            ok = mw.load('bird', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_bird_family_ref:
            ok = mw.load('bird_family', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["BIRD_FAMILY_PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_entry_data:
            ok = mw.load('pdbx', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)

        if args.load_pdbx_core:
            ok = mw.load('pdbx_core', loadType=loadType, inputPathList=inputPathList, styleType=args.document_style,
                         dataSelectors=["PUBLIC_RELEASE"], failedFilePath=failedFilePath,
                         saveInputFileListPath=saveInputFileListPath, pruneDocumentSize=pruneDocumentSize, schemaLevel=schemaLevel)
            okS = loadStatus(mw.getLoadStatus(), cfgOb, readBackCheck=readBackCheck)
        #
        logger.info("Operation completed with status %r " % ok and okS)


if __name__ == '__main__':
    main()



##
# File:    SchemaDefAccess.py
# Author:  J. Westbrook
# Date:    15-Jum-2019
# Version: 0.001 Initial version
#
# Updates:
#  15-Jun-2018 jdw adapted from prior SchemaDefBase() with the static content definitions.
#
##
"""
Schema definition accessor implementation.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging

from rcsb.db.define.SchemaDefAccessBase import SchemaDefAccessBase

logger = logging.getLogger(__name__)


class SchemaDefAccess(SchemaDefAccessBase):
    """ Thin implementation of schema definition access class ...

    """

    def __init__(self, schemaDef, **kwargs):
        super(SchemaDefAccess, self).__init__(schemaDef, **kwargs)



##
# File:    SchemaDefHelper.py
# Author:  J. Westbrook
# Date:    24-May-2018
# Version: 0.001 Initial version
#
# Updates:
#     24-Jul-2018  jdw Make the name conversion method convention specific.
#      4-Nov-2018  jdw add support for excluded attributes
##
"""
Inject additional semantic information into a schema definition applicable to all implementation types.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import re

from rcsb.db.helpers.SchemaDefHelperBase import SchemaDefHelperBase

logger = logging.getLogger(__name__)


class SchemaDefHelper(SchemaDefHelperBase):
    """ Inject additional information into a schema definition. Single source of additional schema semantic content.

    """

    #
    __re0 = re.compile('(database|cell|order|partition|group)$', flags=re.IGNORECASE)
    __re1 = re.compile(r'[-/%[]')
    __re2 = re.compile(r'[\]]')

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: (below)  placeholer

        """
        super(SchemaDefHelper, self).__init__(**kwargs)
        # ----
        #
        self.__cfgOb = kwargs.get('cfgOb', None)
        sectionName = kwargs.get('config_section', 'schemadef_helper')
        self.__cfgD = self.__cfgOb.exportConfig(sectionName=sectionName)
        #
        # ----

    def getConvertNameMethod(self, nameConvention):
        if nameConvention.upper() in ['SQL']:
            return self.__convertNameDefault
        elif nameConvention.upper() in ['ANY', 'DOCUMENT', 'SOLR', 'JSON', 'BSON']:
            return self.__convertNamePunc
        else:
            return self.__convertNameDefault

    def __convertNamePunc(self, name):
        """ Default schema name converter -
        """
        return SchemaDefHelper.__re1.sub('_', SchemaDefHelper.__re2.sub('', name))

    def __convertNameDefault(self, name):
        """ Default schema name converter -
        """
        if SchemaDefHelper.__re0.match(name) or name[0].isdigit():
            name = 'the_' + name
        return SchemaDefHelper.__re1.sub('_', SchemaDefHelper.__re2.sub('', name))

    # @classmethod
    # def xxconvertName(cls, name):
    #    """ Default schema name converter -
    #    """
    #    if cls.__re0.match(name):
    #        name = 'the_' + name
    #    return cls.__re1.sub('_', cls.__re2.sub('', name))

    def getExcluded(self, schemaName):
        '''  For input schema definition, return the list of excluded schema identifiers.

        '''
        includeL = []
        try:
            includeL = [tS.upper() for tS in self.__cfgD['schema_content_filters'][schemaName]['EXCLUDE']]
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return includeL

    def getExcludedAttributes(self, schemaName):
        atExcludeD = {}
        for sn, dL in self.__cfgD['exclude_attributes'].items():
            for d in dL:
                atExcludeD[(d['CATEGORY_NAME'], d['ATTRIBUTE_NAME'])] = sn
        return atExcludeD

    def getIncluded(self, schemaName):
        '''  For input schema definition, return the list of included schema identifiers.

        '''
        excludeL = []
        try:
            excludeL = [tS.upper() for tS in self.__cfgD['schema_content_filters'][schemaName]['INCLUDE']]
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return excludeL

    def getBlockAttributeName(self, schemaName):
        r = None
        try:
            return self.__cfgD['block_attributes'][schemaName]['ATTRIBUTE_NAME']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r

    def getBlockAttributeCifType(self, schemaName):
        r = None
        try:
            return self.__cfgD['block_attributes'][schemaName]['CIF_TYPE_CODE']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r

    def getBlockAttributeMaxWidth(self, schemaName):
        r = None
        try:
            return self.__cfgD['block_attributes'][schemaName]['MAX_WIDTH']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r

    def getBlockAttributeMethod(self, schemaName):
        r = None
        try:
            return self.__cfgD['block_attributes'][schemaName]['METHOD']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r

    def getDatabaseName(self, schemaName):
        r = (None, None)
        try:
            return self.__cfgD['database_names'][schemaName]['NAME']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r

    def getDatabaseVersion(self, schemaName):
        r = (None, None)
        try:
            return self.__cfgD['database_names'][schemaName]['VERSION']
        except Exception as e:
            logger.debug("Schema definition %s failing with %s" % (schemaName, str(e)))
        return r



##
# File:    PdbxLoader.py
# Author:  J. Westbrook
# Date:    14-Mar-2018
# Version: 0.001
#
# Updates:
#     20-Mar-2018 jdw  adding prdcc within chemical component collection
#     21-Mar-2018 jdw  content filtering options added from documents
#     25-Mar-2018 jdw  add support for loading multiple collections for a content type.
#     26-Mar-2018 jdw  improve how successful loads are tracked accross collections -
#      9-Apr-2018 jdw  folding in improved schema path utility apis
#     19-Jun-2018 jdw  integrate with dynamic schema / status object must be unit cardinality.
#     22-Jun-2018 jdw  change collection attribute specification to dot notation
#     22-Jun-2018 jdw  separate cases where the loading success can be easily mapped to the source data object.
#     24-Jun-2018 jdw  rename and specialize function
#     14-Jul-2018 jdw  add methods to return data exchange status objects for load operations
#     25-Jul-2018 jdw  fixed bazaar bad function references blocking failure handling
#     25-Jul-2018 jdw  restore pruning operations
#     14-Aug-2018 jdw  primaryIndexD from self.__schU.getSchemaInfo(schemaName) updated to list and i
#                      in __createCollection(self, dbName, collectionName, indexAttributeNames=None) make indexAttributeNames a list
#     10-Sep-2018 jdw  Adjust error handling and reporting across multiple collections
#     24-Oct-2018 jdw  update for new configuration organization
#     11-Nov-2018 jdw  add DrugBank and CCDC mapping path details.
#     21-Nov-2018 jdw  add addDocumentPrivateAttributes(dList, collectionName) to inject private document keys
#
##
"""
Worker methods for loading primary data content following mapping conventions in external schema definitions.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
import operator
import sys
import time

import bson

from rcsb.db.define.DictMethodRunner import DictMethodRunner
from rcsb.db.mongo.Connection import Connection
from rcsb.db.mongo.MongoDbUtil import MongoDbUtil
from rcsb.db.processors.DataExchangeStatus import DataExchangeStatus
from rcsb.db.processors.DataTransformFactory import DataTransformFactory
from rcsb.db.processors.SchemaDefDataPrep import SchemaDefDataPrep
from rcsb.db.utils.SchemaDefUtil import SchemaDefUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)


class PdbxLoader(object):

    def __init__(self, cfgOb, cfgSectionName='site_info', resourceName="MONGO_DB", numProc=4, chunkSize=15,
                 fileLimit=None, verbose=False, readBackCheck=False, maxStepLength=2000, workPath=None):
        """  Worker methods for loading primary data content following mapping conventions in external schema definitions.

        Args:
            cfgOb (TYPE): Description
            resourceName (str, optional): Description
            numProc (int, optional): Description
            chunkSize (int, optional): Description
            fileLimit (None, optional): Description
            verbose (bool, optional): Description
            readBackCheck (bool, optional): Description
            maxStepLength (int, optional): Description
            workPath (None, optional): Description
        """
        self.__verbose = verbose
        #
        # Limit the load length of each file type for testing  -  Set to None to remove -
        self.__fileLimit = fileLimit
        self.__maxStepLength = maxStepLength
        #
        # Controls for multiprocessing execution -
        self.__numProc = numProc
        self.__chunkSize = chunkSize
        #
        self.__cfgOb = cfgOb
        self.__resourceName = resourceName
        #
        self.__readBackCheck = readBackCheck
        self.__workPath = workPath
        self.__mpFormat = '[%(levelname)s] %(asctime)s %(processName)s-%(module)s.%(funcName)s: %(message)s'
        #
        self.__schU = SchemaDefUtil(cfgOb=self.__cfgOb, numProc=self.__numProc, fileLimit=self.__fileLimit)
        #
        self.__statusList = []
        #
        sectionName = cfgSectionName
        pathPdbxDictionaryFile = self.__cfgOb.getPath('PDBX_DICT_LOCATOR', sectionName=sectionName)
        pathRcsbDictionaryFile = self.__cfgOb.getPath('RCSB_DICT_LOCATOR', sectionName=sectionName)
        #
        pathDrugBankMappingFile = self.__cfgOb.getPath('DRUGBANK_MAPPING_LOCATOR', sectionName=sectionName)
        pathCsdModelMappingFile = self.__cfgOb.getPath('CCDC_MAPPING_LOCATOR', sectionName=sectionName)
        #
        dH = self.__cfgOb.getHelper('DICT_METHOD_HELPER_MODULE', sectionName=sectionName,
                                    drugBankMappingFilePath=pathDrugBankMappingFile,
                                    workPath=self.__workPath,
                                    csdModelMappingFilePath=pathCsdModelMappingFile)
        self.__dmh = DictMethodRunner(dictLocators=[pathPdbxDictionaryFile, pathRcsbDictionaryFile], methodHelper=dH)

    def load(self, schemaName, collectionLoadList=None, loadType='full', inputPathList=None, styleType='rowwise_by_name', dataSelectors=None,
             failedFilePath=None, saveInputFileListPath=None, pruneDocumentSize=None, logSize=False,
             schemaLevel='min', locatorKey='rcsb_load_status.locator'):
        """Driver method for loading PDBx/mmCIF content into the Mongo document store.

        Args:
            schemaName (str): A content schema (e.g. 'bird','bird_family','bird_chem_comp', chem_comp', 'pdbx', 'pdbx_core')
            collectionLoadList (list, optional): list of collection names in this schema to load (default is load all collections)
            loadType (str, optional): mode of loading 'full' (bulk delete then bulk insert) or 'replace'
            inputPathList (list, optional): Data file path list (if not provided the full repository will be scanned)
            styleType (str, optional): one of 'rowwise_by_name', 'columnwise_by_name', 'rowwise_no_name', 'rowwise_by_name_with_cardinality'
            dataSelectors (list, optional): selector names defined for this schema (e.g. PUBLIC_RELEASE)
            failedFilePath (str, optional): Path to hold file paths for load failures
            saveInputFileListPath (list, optional): List of files
            pruneDocumentSize (bool, optional): iteratively remove large elements from a collection to satisfy size limits
            logSize (bool, optional): Compute and log bson serialized object size
            schemaLevel (str, optional): Completeness of json/bson metadata schema bound to each collection (e.g. 'min', 'full' or None)
            locatorKey (str, optional): Key identifier in document content storing the data file path (url)


        Returns:
            bool: True on success or False otherwise

        """
        try:
            #
            self.__statusList = []
            desp = DataExchangeStatus()
            statusStartTimestamp = desp.setStartTime()
            #
            startTime = self.__begin(message="loading operation")
            #
            pathList = self.__schU.getPathList(contentType=schemaName, inputPathList=inputPathList)
            logger.debug("Path list length %d" % len(pathList))
            #
            if saveInputFileListPath:
                self.__writePathList(saveInputFileListPath, pathList)
                logger.info("Saving %d paths in %s" % (len(pathList), saveInputFileListPath))
            #
            filterType = "drop-empty-attributes|drop-empty-tables|skip-max-width|assign-dates|convert-iterables|normalize-enums"
            if styleType in ["columnwise_by_name", "rowwise_no_name"]:
                filterType = "drop-empty-tables|skip-max-width|assign-dates|convert-iterables|normalize-enums"
            #
            optD = {}
            optD['schemaName'] = schemaName
            optD['styleType'] = styleType
            optD['filterType'] = filterType
            optD['readBackCheck'] = self.__readBackCheck
            optD['dataSelectors'] = dataSelectors
            optD['loadType'] = loadType
            optD['logSize'] = logSize
            optD['pruneDocumentSize'] = pruneDocumentSize
            optD['locatorKey'] = locatorKey
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            #

            numProc = self.__numProc
            chunkSize = self.__chunkSize if pathList and self.__chunkSize < len(pathList) else 0
            #
            sd, dbName, fullCollectionNameList, primaryIndexD = self.__schU.getSchemaInfo(schemaName)

            collectionNameList = collectionLoadList if collectionLoadList else fullCollectionNameList

            for collectionName in collectionNameList:
                if loadType == 'full':
                    self.__removeCollection(dbName, collectionName)
                    indAtL = primaryIndexD[collectionName] if collectionName in primaryIndexD else None
                    bsonSchema = None
                    if schemaLevel and schemaLevel in ['min', 'full']:
                        bsonSchema = self.__schU.getJsonSchema(collectionName, level=schemaLevel)
                    ok = self.__createCollection(dbName, collectionName, indAtL, bsonSchema=bsonSchema)
                    logger.debug("Collection create return status %r" % ok)

            #
            dtf = DataTransformFactory(schemaDefAccessObj=sd, filterType=filterType)
            optD['schemaDefAccess'] = sd
            optD['dataTransformFactory'] = dtf
            optD['collectionNameList'] = collectionNameList
            optD['dbName'] = dbName
            # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
            numPaths = len(pathList)
            logger.debug("Processing %d total paths" % numPaths)
            numProc = min(numProc, numPaths)
            maxStepLength = self.__maxStepLength
            if numPaths > maxStepLength:
                numLists = int(numPaths / maxStepLength)
                subLists = [pathList[i::numLists] for i in range(numLists)]
            else:
                subLists = [pathList]
            #
            if subLists and len(subLists) > 0:
                logger.info("Starting loadType %s with numProc %d outer subtask count %d subtask length %d" % (loadType, numProc, len(subLists), len(subLists[0])))
            #
            failList = []
            for ii, subList in enumerate(subLists):
                logger.info("Running outer subtask %d of %d length %d" % (ii + 1, len(subLists), len(subList)))
                #
                mpu = MultiProcUtil(verbose=True)
                mpu.setWorkingDir(self.__workPath)
                mpu.setOptions(optionsD=optD)
                mpu.set(workerObj=self, workerMethod="loadWorker")
                ok, failListT, _, _ = mpu.runMulti(dataList=subList, numProc=numProc, numResults=1, chunkSize=chunkSize)
                failList.extend(failListT)
            failList = list(set(failList))
            logger.debug("Failing path list %r" % failList)
            #
            if failedFilePath and len(failList) > 0:
                wOk = self.__writePathList(failedFilePath, failList)
                logger.info("Writing failure path %s length %d status %r" % (failedFilePath, len(failList), wOk))
            #
            ok = len(failList) == 0
            self.__end(startTime, "Loading operation with status " + str(ok))
            #
            # Create the status objects for the current operations
            # ----
            sFlag = 'Y' if ok else 'N'
            for collectionName in collectionNameList:
                desp.setStartTime(tS=statusStartTimestamp)
                desp.setObject(dbName, collectionName)
                desp.setStatus(updateId=None, successFlag=sFlag)
                desp.setEndTime()
                self.__statusList.append(desp.getStatus())
            #
            logger.info("Load operation success status %r failed paths %d of %d" % (ok, len(failList), numPaths))
            return ok
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return False

    def getLoadStatus(self):
        return self.__statusList

    def loadWorker(self, dataList, procName, optionsD, workingDir):
        """ Multi-proc worker method for MongoDb loading -

        sucessList,resultList,diagList=workerFunc(runList=nextList,procName, optionsD, workingDir)
        """
        try:
            startTime = self.__begin(message=procName)
            # Recover common options
            styleType = optionsD['styleType']
            filterType = optionsD['filterType']
            readBackCheck = optionsD['readBackCheck']
            logSize = 'logSize' in optionsD and optionsD['logSize']
            dataSelectors = optionsD['dataSelectors']
            loadType = optionsD['loadType']
            schemaName = optionsD['schemaName']
            pruneDocumentSize = optionsD['pruneDocumentSize']
            locatorKey = optionsD['locatorKey']
            #
            sd = optionsD['schemaDefAccess']
            dtf = optionsD['dataTransformFactory']
            collectionNameList = optionsD['collectionNameList']
            dbName = optionsD['dbName']

            #
            sdp = SchemaDefDataPrep(schemaDefAccessObj=sd, dtObj=dtf, workPath=workingDir, verbose=self.__verbose)
            containerList = sdp.getContainerList(dataList, filterType=filterType)
            #
            # Apply dynamic methods for generated content here -
            #
            readSuccessPathList = []
            for container in containerList:
                if self.__dmh:
                    self.__dmh.apply(container)
                readSuccessPathList.append(container.getProp('locator'))
            #
            readFailPathList = list(set(dataList) - set(readSuccessPathList))
            #
            if readFailPathList:
                logger.info("Read failures readFailPathList %r" % readFailPathList)
            #
            logger.debug("%s container count %d" % (procName, len(containerList)))
            #
            logger.debug("%s schemaName %s dbName %s collectionNameList %s pathlist %d containerList %d" %
                         (procName, schemaName, dbName, collectionNameList, len(dataList), len(containerList)))
            fullSuccessPathList = []
            fullFailedPathList = readFailPathList
            for collectionName in collectionNameList:
                successPathList = []
                failedPathList = []
                ok = True
                # ---------------- - ---------------- - ---------------- - ---------------- - ---------------- -
                tableIdExcludeList = sd.getCollectionExcluded(collectionName)
                tableIdIncludeList = sd.getCollectionSelected(collectionName)
                sliceFilter = sd.getCollectionSliceFilter(collectionName)
                sdp.setSchemaIdExcludeList(tableIdExcludeList)
                sdp.setSchemaIdIncludeList(tableIdIncludeList)
                #
                logger.debug("%s schemaName %s dbName %s collectionName %s slice filter %s" % (procName, schemaName, dbName, collectionName, sliceFilter))
                logger.debug("%s schemaName %s include list %r" % (procName, schemaName, tableIdIncludeList))
                logger.debug("%s schemaName %s exclude list %r" % (procName, schemaName, tableIdExcludeList))

                #
                dList, containerNameList, rejectPathList = sdp.processDocuments(containerList, styleType=styleType, filterType=filterType,
                                                                                dataSelectors=dataSelectors, sliceFilter=sliceFilter)
                #
                # Get the unique paths for the rejected/filtered container list - (rejections are NOT treated as failures)
                #
                rejectPathList = list(set(rejectPathList))
                #
                logger.debug("%s schemaName %s dbName %s collectionName %s slice filter %s num containers %d len data %d num rejects %d" %
                             (procName, schemaName, dbName, collectionName, sliceFilter, len(containerNameList), len(dList), len(rejectPathList)))
                #
                if logSize:
                    maxDocumentMegaBytes = -1
                    for tD, cN in zip(dList, containerNameList):
                        # documentMegaBytes = float(sys.getsizeof(pickle.dumps(tD, protocol=0))) / 1000000.0
                        documentMegaBytes = float(sys.getsizeof(bson.BSON.encode(tD))) / 1000000.0
                        logger.debug("%s Document %s  %.4f MB" % (procName, cN, documentMegaBytes))
                        maxDocumentMegaBytes = max(maxDocumentMegaBytes, documentMegaBytes)
                        if documentMegaBytes > 15.8:
                            logger.info("Large document %s  %.4f MB" % (cN, documentMegaBytes))
                    logger.debug("%s maximum document size loaded %.4f MB" % (procName, maxDocumentMegaBytes))
                #
                #  Get the [scbemaId.atId,...] holding the natural document Id
                #
                docIdL = sd.getDocumentKeyAttributeNames(collectionName)
                #
                #  Add any private document keys -
                dList = sdp.addDocumentPrivateAttributes(dList, collectionName)

                logger.debug("%s docIdL %r collectionName %r" % (procName, docIdL, collectionName))
                #
                if dList:
                    ok, successPathList, failedPathList = self.__loadDocuments(dbName, collectionName, dList, docIdL,
                                                                               loadType=loadType, locatorKey=locatorKey,
                                                                               readBackCheck=readBackCheck, pruneDocumentSize=pruneDocumentSize)
                #
                successPathList.extend(list(set(rejectPathList)))
                fullSuccessPathList.extend(list(set(successPathList)))
                fullFailedPathList.extend(list(set(failedPathList)))
                logger.info("%s %s/%s document list %d successes %d  failures %d filtered/rejected %d" %
                            (procName, dbName, collectionName, len(dList), len(successPathList), len(failedPathList), len(rejectPathList)))
            #
            logger.debug("fullSuccessPathList %r" % fullSuccessPathList)
            logger.debug("fullFailedPathList  %r" % fullFailedPathList)
            retList = list(set(dataList) - set(fullFailedPathList))
            #
            logger.debug("Length input path list %d length filtered list %d returning path list %d" % (len(dataList), len(rejectPathList), len(retList)))
            #
            logger.info("%s %s %r returning success path list %d full failed path list %d " % (procName, dbName, collectionNameList,
                                                                                               len(set(retList)), len(set(fullFailedPathList))))
            self.__end(startTime, procName + " with status " + str(ok))
            return retList, [], []

        except Exception as e:
            # logger.error("Failing for dataList %r" % dataList)
            logger.exception("Failing with %s" % str(e))

        return [], [], []

    # -------------- -------------- -------------- -------------- -------------- -------------- --------------
    #                                        ---  Supporting code follows ---
    #
    def __writePathList(self, filePath, pathList):
        try:
            with open(filePath, 'w') as ofh:
                for pth in pathList:
                    ofh.write("%s\n" % pth)
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def __begin(self, message=""):
        startTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        logger.debug("Starting %s at %s" % (message, ts))
        return startTime

    def __end(self, startTime, message=""):
        endTime = time.time()
        ts = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
        delta = endTime - startTime
        logger.debug("Completed %s at %s (%.4f seconds)" % (message, ts, delta))

    def __createCollection(self, dbName, collectionName, indexAttributeNames=None, bsonSchema=None):
        """Create database and collection and optionally a primary index -
        """
        try:
            logger.debug("Create database %s collection %s" % (dbName, collectionName))
            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                ok1 = mg.createCollection(dbName, collectionName, bsonSchema=bsonSchema)
                ok2 = mg.databaseExists(dbName)
                ok3 = mg.collectionExists(dbName, collectionName)
                okI = True
                if indexAttributeNames:
                    okI = mg.createIndex(dbName, collectionName, indexAttributeNames, indexName="primary", indexType="DESCENDING", uniqueFlag=False)

            return ok1 and ok2 and ok3 and okI
            #
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def __removeCollection(self, dbName, collectionName):
        """Drop collection within database

        """
        try:
            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                #
                logger.debug("Remove collection database %s collection %s" % (dbName, collectionName))
                logger.debug("Starting databases = %r" % mg.getDatabaseNames())
                logger.debug("Starting collections = %r" % mg.getCollectionNames(dbName))
                ok = mg.dropCollection(dbName, collectionName)
                logger.debug("Databases = %r" % mg.getDatabaseNames())
                logger.debug("Post drop collections = %r" % mg.getCollectionNames(dbName))
                ok = mg.collectionExists(dbName, collectionName)
                logger.debug("Post drop collections = %r" % mg.getCollectionNames(dbName))
            return ok
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
        return False

    def __pruneBySize(self, dList, limitMB=15.9):
        """ For the input list of objects (dictionaries).objects
            Return a pruned list satisfying the input total object size limit -

        """
        oL = []
        try:
            for d in dList:
                sD = {}
                sumMB = 0.0
                for ky in d:
                    dMB = float(sys.getsizeof(bson.BSON.encode({ky: d[ky]}))) / 1000000.0
                    sumMB += dMB
                    sD[ky] = dMB
                if sumMB < limitMB:
                    oL.append(d)
                    continue
                #
                sorted_sD = sorted(sD.items(), key=operator.itemgetter(1))
                prunedSum = 0.0
                for ky, sMB in sorted_sD:
                    prunedSum += sMB
                    if prunedSum > limitMB:
                        d.pop(ky, None)
                        logger.debug("Pruning ky %s size(MB) %.2f" % (ky, sMB))
                oL.append(d)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            #
        logger.debug("Pruning returns document list length %d" % len(dList))
        return oL

    def __loadDocuments(self, dbName, collectionName, dList, docIdL, loadType='full', locatorKey=None, readBackCheck=False, pruneDocumentSize=None):
        #
        # Load database/collection with input document list -
        #
        rIdL = []
        indD = {}
        failList = []
        fullSuccessValueList = [self.__getKeyValue(d, locatorKey) for d in dList]
        logger.debug("fullSuccessValueList length %d" % len(fullSuccessValueList))

        logger.debug("dbName %s collectionName %s docIdL %r locatorKey %r" % (dbName, collectionName, docIdL, locatorKey))
        #
        try:
            for ii, d in enumerate(dList):
                dIdTup = self.__getKeyValues(d, docIdL)
                indD[dIdTup] = ii
        except Exception as e:
            logger.exception("Failing ii %d d %r with %s" % (ii, d, str(e)))

        try:
            with Connection(cfgOb=self.__cfgOb, resourceName=self.__resourceName) as client:
                mg = MongoDbUtil(client)
                #
                keyNames = docIdL
                if loadType == 'replace':
                    dTupL = mg.deleteList(dbName, collectionName, dList, keyNames)
                    logger.debug("Deleted document status %r" % dTupL)
                if pruneDocumentSize:
                    dList = self.__pruneBySize(dList, limitMB=pruneDocumentSize)
                #
                rIdL.extend(mg.insertList(dbName, collectionName, dList, keyNames=keyNames, salvage=True))
                logger.debug("-- InsertList returns rIdL length %r of %r" % (len(rIdL), len(dList)))
                # ---
                #  If there is a failure then determine the specific successes and failures -
                #
                successList = fullSuccessValueList
                if len(rIdL) != len(dList):
                    try:
                        sList = []
                        for rId in rIdL:
                            #
                            rObj = mg.fetchOne(dbName, collectionName, '_id', rId)
                            dIdTup = self.__getKeyValues(rObj, docIdL)
                            jj = indD[dIdTup]
                            sList.append(self.__getKeyValue(dList[jj], locatorKey))
                        #
                        failList = list(set(fullSuccessValueList) - set(sList))
                        successList = list(set(sList))
                    except Exception as e:
                        logger.exception("Failing with %s" % str(e))
                #
                if readBackCheck:
                    #
                    # Note that objects in dList are mutated by the insert operation with the additional key '_id',
                    # hence, it is possible to compare the fetched object with the input object.
                    #
                    rbStatus = True
                    for ii, rId in enumerate(rIdL):
                        rObj = mg.fetchOne(dbName, collectionName, '_id', rId)
                        dIdTup = self.__getKeyValues(rObj, docIdL)
                        jj = indD[dIdTup]
                        if (rObj != dList[jj]):
                            rbStatus = False
                            break
                #
                if readBackCheck and not rbStatus:
                    return False, successList, failList
                #
            return len(rIdL) == len(dList), successList, failList
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        return False, [], fullSuccessValueList

    def __getKeyValues(self, dct, keyNames):
        """Return the tuple of values of corresponding to the input dictionary key names expressed in dot notation.

        Args:
            dct (dict): source dictionary object (nested)
            keyNames (list): list of dictionary keys in dot notatoin

        Returns:
            tuple: tuple of values corresponding to the input key names

        """
        rL = []
        try:
            for keyName in keyNames:
                rL.append(self.__getKeyValue(dct, keyName))
        except Exception as e:
            logger.exception("Failing for key names %r with %s" % (keyNames, str(e)))

        return tuple(rL)

    def __getKeyValue(self, dct, keyName):
        """  Return the value of the corresponding key expressed in dot notation in the input dictionary object (nested).
        """
        try:
            kys = keyName.split('.')
            for key in kys:
                try:
                    dct = dct[key]
                except KeyError:
                    return None
            return dct
        except Exception as e:
            logger.exception("Failing for key %r with %s" % (keyName, str(e)))

        return None



##
# File:    DictInfoHelper.py
# Author:  J. Westbrook
# Date:    18-May-2018
# Version: 0.001 Initial version
#
# Updates:
#  23-May-2018  jdw revise dependencies for helper function.  Change method api's
#  27-May-2018  jdw add attribrute level filters
#   1-Jun-2018  jdw add bridging class DataTransformInfo for attribute filters
#  13-Jun-2018  jdw add content classes to cover former base table feature
#  15-Jun-2018  jdw add support for alternative iterable delimiters
#  24-Jul-2018  jdw fix logic in processing _itemTransformers data
#   6-Aug-2018  jdw add slice parent item definitions in terms of parent items
#   8-Aug-2018  jdw add slice parent conditional filter definitions that could be applied to the parent data category,
#  10-Aug-2018  jdw add slice category and cardinality extras
#  18-Aug-2018  jdw add schema pdbx_core analogous to pdbx removing the block attribute.
#   7-Sep-2018  jdw add generated content classes for core schemas
#  10-Sep-2018  jdw add iterable details for semicolon separated text data
#  11-Sep-2018  jdw adjust slice cardinality constraints for entity and assembly identifier categories.
#  30-Sep-2018  jdw add source and host organism categories
#   2-Oct-2018  jdw add repository_holdings and sequence_cluster content types and associated category content.
##
"""
This helper class supplements dictionary information as required for schema production.
I other words, the additional content conferred here would best be incorporated as part
of standard data definitions at some future point.

This is the single source of additional dictionary specific semantic content and is
design to avoid scattering dictionary semantic content haphazardly throughout the
code base.

All data accessors and structures here refer to dictionary category and attribute names.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.db.helpers.DictInfoHelperBase import DictInfoHelperBase
from rcsb.db.processors.DataTransformFactory import DataTransformInfo

logger = logging.getLogger(__name__)


class DictInfoHelper(DictInfoHelperBase):
    """ Supplements dictionary information as required for schema production.

    """

    def __init__(self, **kwargs):
        """
        Args:
            **kwargs: (below)
            dictPath (str, optional): path to the current dictioonary text
            dictSubset (str, optional): name of dictionary content subset - alias for schema name


            Add - Include exclude filters on dictionary content -


        """
        super(DictInfoHelper, self).__init__(**kwargs)
        # ----
        #
        self.__cfgOb = kwargs.get('cfgOb', None)
        sectionName = kwargs.get('config_section', 'dictionary_helper')
        self.__cfgD = self.__cfgOb.exportConfig(sectionName=sectionName)
        #
        # ----
        self._dictSubset = kwargs.get('dictSubset', None)
        self._dictPath = kwargs.get("dictPath", None)
        #
        self.__dti = DataTransformInfo()
        self.__itD = self.__getItemTransformD()
        #
        self.__categoryClasses = self.__getCategoryContentClasses()
        self.__attributeClasses = self.__getAttributeContentClasses()

    def getDelimiter(self, categoryName, attributeName, default=','):
        for tD in self.__cfgD['iterable_delimiters']:
            if tD['CATEGORY_NAME'] == categoryName and tD['ATTRIBUTE_NAME'] == attributeName:
                return tD['DELIMITER']
        return default

    def getCategoryContentClasses(self, dictSubset):
        try:
            rD = {}
            for catName, cDL in self.__categoryClasses.items():
                for cD in cDL:
                    if cD['DICT_SUBSET'] == dictSubset:
                        if catName not in rD:
                            rD[catName] = []
                        rD[catName].append(cD['CONTENT_CLASS'])

        except Exception as e:
            logger.debug("Failing with %s" % str(e))
        return rD

    def getAttributeContentClasses(self, dictSubset):
        try:
            rD = {}
            for (catName, atName), cDL in self.__attributeClasses.items():
                for cD in cDL:
                    if cD['DICT_SUBSET'] == dictSubset:
                        if (catName, atName) not in rD:
                            rD[(catName, atName)] = []
                        rD[(catName, atName)].append(cD['CONTENT_CLASS'])

        except Exception as e:
            logger.debug("Failing with %s" % str(e))
        return rD

    def __getCategoryContentClasses(self):
        classD = {}
        try:
            for cTup, cDL in self.__cfgD['content_classes'].items():
                for cD in cDL:
                    if cD['CATEGORY_NAME'] not in classD:
                        classD[cD['CATEGORY_NAME']] = []
                    classD[cD['CATEGORY_NAME']].append({'CONTENT_CLASS': cTup[0], 'DICT_SUBSET': cTup[1]})
        except Exception as e:
            logger.debug("Failing with %s" % str(e))
        return classD

    def __getAttributeContentClasses(self):
        classD = {}
        try:
            for cTup, cDL in self.__cfgD['content_classes'].items():
                for cD in cDL:
                    catName = cD['CATEGORY_NAME']
                    for atName in cD['ATTRIBUTE_NAME_LIST']:
                        if (catName, atName) not in classD:
                            classD[(catName, atName)] = []
                        classD[(catName, atName)].append({'CONTENT_CLASS': cTup[0], 'DICT_SUBSET': cTup[1]})
        except Exception as e:
            logger.debug("Failing with %s" % str(e))
        return classD

    def __getItemTransformD(self):
        itD = {}
        for f, dL in self.__cfgD['item_transformers'].items():
            logger.debug("Verify transform method %r" % f)
            if self.__dti.isImplemented(f):
                for d in dL:
                    if (d['CATEGORY_NAME'], d['ATTRIBUTE_NAME']) not in itD:
                        itD[(d['CATEGORY_NAME'], d['ATTRIBUTE_NAME'])] = []
                    itD[(d['CATEGORY_NAME'], d['ATTRIBUTE_NAME'])].append(f)

        return itD

    def getDictPath(self):
        return self._dictPath

    def getDictSubSet(self):
        return self._dictSubset

    def getTransformItems(self, transformName):
        """ Return the list of items subject to the input attribute filter.

            _itemTransformers{<filterName> : [{'CATEGORY_NAME':..., 'ATTRIBUTE_NAME': ... },{}]
        """
        try:
            if self.__dti.isImplemented(transformName):
                return self.__cfgD['item_transformers'][transformName]
        except Exception:
            return []

    def getItemTransforms(self, categoryName, attributeName):
        """ Return the list of transforms to be applied to the input item (categoryName, attributeName).
        """
        try:
            return self.__itD[(categoryName, attributeName)]
        except Exception:
            return []

    def getItemTransformD(self):
        """ Return the dictionary of transforms to be applied to the input item (categoryName, attributeName).
        """
        try:
            return self.__itD
        except Exception:
            return {}

    def getCardinalityCategoryExtras(self):
        return self.__cfgD['cardinality_category_extras']

    def getCardinalityKeyItem(self, dictSubset):
        """ Identify the parent item for the dictionary subset that can be used to
            identify child categories with unity cardinality.   That is, logically containing
            a single data row in any instance.

        """
        try:
            return self.__cfgD['cardinality_parent_items'][dictSubset]
        except Exception:
            pass
        return {'CATEGORY_NAME': None, 'ATTRIBUTE_NAME': None}

    def getTypeCodes(self, kind):
        """
        """
        try:
            return self.__cfgD['type_code_classes'][kind]
        except Exception as e:
            logger.exception("Failing for kind %r with %s" % (kind, str(e)))
            pass
        return []

    def getQueryStrings(self, kind):
        try:
            return self.__cfgD['query_string_selectors'][kind]
        except Exception:
            pass

        return []

    def getSelectionFilter(self, dictSubset, kind):
        """  Interim api for selection filters defined in terms of dictionary category and attributes name and their values.

        """
        try:
            return self.__cfgD['selection_filters'][(kind, dictSubset)]
        except Exception:
            pass
        return []

    def getSelectionFiltersBySubset(self, dictSubset):
        """  Interim api for selection filters for a particular dictionary subset.

        """
        try:
            return {kind: v for (kind, dS), v in self.__cfgD['selection_filters'].items() if dS == dictSubset}
        except Exception:
            pass
        return {}

    def getContentClass(self, dictSubset, kind):
        """  Interim api for special category classes.

        """
        try:
            return self.__cfgD['special_content'][(kind, dictSubset)]
        except Exception:
            pass
        return []

    def getContentClassBySubset(self, dictSubset):
        """  Interim api for special category classes.

        """
        try:
            return {kind: v for (kind, dS), v in self.__cfgD['special_content'].items() if dS == dictSubset}
        except Exception:
            pass
        return {}

    def getSliceParentItems(self, dictSubset, kind):
        """  Interim api for slice parent itens defined in terms of dictionary category and attributes name and their values.

        """
        try:
            return self.__cfgD['slice_parent_items'][(kind, dictSubset)]
        except Exception:
            pass
        return []

    def getSliceParentsBySubset(self, dictSubset):
        """  Interim api for slice parent items for a particular dictionary subset.

        """
        try:
            return {kind: v for (kind, dS), v in self.__cfgD['slice_parent_items'].items() if dS == dictSubset}
        except Exception:
            pass
        return {}

    def getSliceParentFilters(self, dictSubset, kind):
        """  Interim api for slice parent condition filters defined in terms of dictionary category and attributes name and their values.

        """
        try:
            return self.__cfgD['slice_parent_filters'][(kind, dictSubset)]
        except Exception:
            pass
        return []

    def getSliceParentFiltersBySubset(self, dictSubset):
        """  Interim api for slice parent condition filters for a particular dictionary subset.

        """
        try:
            return {kind: v for (kind, dS), v in self.__cfgD['slice_parent_filters'].items() if dS == dictSubset}
        except Exception:
            pass
        return {}

    def getSliceCardinalityCategoryExtras(self, dictSubset, kind):
        try:
            return self.__cfgD['slice_cardinality_category_extras'][(kind, dictSubset)]
        except Exception:
            return []

    def getSliceCategoryExtras(self, dictSubset, kind):
        try:
            return self.__cfgD['slice_category_extras'][(kind, dictSubset)]
        except Exception:
            return []



##
# File:    DataTransformFactory.py
# Author:  J. Westbrook
# Date:    10-Apr-2018
#
#
# Updates:
#
#  4-Sep-2018 jdw add enumeration normalization (in progress)
##
"""
Factory for functional elements of the transformations between input data and
and loadable data using specifications from the schema map definition.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import collections
import logging
import re
from functools import reduce

import dateutil.parser

import pytz

TrfValue = collections.namedtuple('TrfValue', 'value, atId, origLength, isNull')

logger = logging.getLogger(__name__)


class DataTransformInfo(object):
    """ Map transformation attribute filter names to data transformation filter implementations.

    """

    def __init__(self):
        # mapD {<external attribute filter name>: <implementation names>, ...}
        self.__mapD = {'STRIP_WS': 'STRIP_WS'}

    def isImplemented(self, attributeFilter):
        ok = False
        try:
            if attributeFilter in self.__mapD:
                ok = True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        logger.debug("Filter %r status %r" % (attributeFilter, ok))
        return ok

    def getTransformFilterName(self, attributeFilter):
        try:
            return self.__mapD[attributeFilter]
        except Exception:
            pass
        return None


class DataTransformFactory(object):
    """Factory for functional elements of the transformations between input data and
        and loadable data using specifications from the schema map definition.

        input string value -> (Null Handling -> cast type ->)

        input string iterable value ->


    """

    def __init__(self, schemaDefAccessObj, filterType):
        self.__sD = schemaDefAccessObj
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)
        logger.debug("filterType %r" % filterType)
        logger.debug("Schema database %r" % self.__sD.getDatabaseName())
        self.__FLAGS = {}
        self.__FLAGS['dropEmpty'] = 'drop-empty-attributes' in filterType
        self.__FLAGS['skipMaxWidth'] = 'skip-max-width' in filterType
        self.__FLAGS['assignDates'] = 'assign-dates' in filterType
        self.__FLAGS['convertIterables'] = 'convert-iterables' in filterType
        self.__FLAGS['normalizeEnums'] = 'normalize-enums' in filterType
        self.__FLAGS['normalizeDates'] = True
        logger.debug("FLAGS settings are %r" % self.__FLAGS)
        #
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)
        self.__dti = DataTransformInfo()
        self.__dT = self.__build()

    def __build(self):
        """ Internal method that stores transformations for each table so that these may
            be later repeatedly applied to a data stream.

           fD[tableId]'atFuncD'] -> {atId1: [f1,f2,.. ], atId2: [f1,f2,...], ...}
        """
        fD = {}
        for tableId in self.__sD.getSchemaIdList():
            tD = {}
            tObj = self.__sD.getSchemaObject(tableId)
            dt = DataTransform(tObj)
            aD = {}
            for atId in tObj.getAttributeIdList():
                if tObj.isOtherAttributeType(atId):
                    # skip attributes with no mapping correspondence
                    continue
                #
                aD[atId] = []
                #
                if self.__FLAGS['convertIterables'] and tObj.isIterable(atId):
                    if tObj.isAttributeStringType(atId):
                        aD[atId].append(dt.castIterableString)
                    elif tObj.isAttributeIntegerType(atId):
                        aD[atId].append(dt.castIterableInteger)
                    elif tObj.isAttributeFloatType(atId):
                        aD[atId].append(dt.castIterableFloat)
                #
                elif tObj.isAttributeStringType(atId):
                    aD[atId].append(dt.castString)
                    if not self.__FLAGS['skipMaxWidth']:
                        aD[atId].append(dt.truncateString)
                    #
                    for ft in tObj.getAttributeFilterTypes(atId):
                        if self.__dti.getTransformFilterName(ft) == "STRIP_WS":
                            aD[atId].append(dt.stripWhiteSpace)
                elif tObj.isAttributeIntegerType(atId):
                    aD[atId].append(dt.castInteger)
                elif tObj.isAttributeFloatType(atId):
                    aD[atId].append(dt.castFloat)
                elif self.__FLAGS['assignDates'] and tObj.isAttributeDateType(atId):
                    aD[atId].append(dt.castDateToObj)
                elif self.__FLAGS['normalizeDates'] and tObj.isAttributeDateType(atId):
                    if tObj.getAttributeType(atId).lower() == 'datetime':
                        aD[atId].append(dt.castDateTimeToIsoDate)
                    elif tObj.getAttributeType(atId).lower() == 'date':
                        aD[atId].append(dt.castDateToIsoDate)
                else:
                    aD[atId].append(dt.castString)
                #
                if self.__FLAGS['normalizeEnums'] and tObj.isEnumerated(atId):
                    logger.debug("Normalizing enums for %s %s" % (tableId, atId))
                    aD[atId].append(dt.normalizeEnum)
            #
            # Transformation functions keyed by attribute 'name'
            tD['atIdD'] = tObj.getMapAttributeIdDict()
            tD['atNameD'] = tObj.getMapAttributeNameDict()
            tD['atNullValues'] = tObj.getAppNullValueDict()
            tD['atFuncD'] = {tD['atIdD'][k]: v for k, v in aD.items()}
            #
            fD[tableId] = tD
        ##
        return fD

    def get(self, tableId):
        try:
            return self.__dT[tableId]
        except Exception as e:
            logger.error("Missing table %r with error %s" % (tableId, str(e)))
        return {}

    def processRecord(self, tableId, row, attributeNameList):
        """
            Input row data (list) ordered according to the input attribute names list.

            Processing respects various null handling policies.


            return   d[atId]=rowdata for the input row list

        """
        # get the transform object for the current table
        dT = self.get(tableId)
        #
        atName = None
        try:
            d = {} if self.__FLAGS['dropEmpty'] else {k: v for k, v in dT['atNullValues'].items()}
            for ii, atName in enumerate(attributeNameList):
                if atName not in dT['atNameD']:
                    continue
                #
                nullFlag = False if row[ii] else True

                # Apply list of functions on an initial value (i.e. TrfValue for the ii(th) element of the row.
                vT = reduce(lambda x, y: y(x), dT['atFuncD'][atName], TrfValue(row[ii], dT['atNameD'][atName], 0, nullFlag))
                if self.__FLAGS['dropEmpty'] and vT.isNull:
                    continue
                d[dT['atNameD'][atName]] = vT.value
        except Exception as e:
            logger.exception("Failing with %s for table %s atName %s" % (str(e), tableId, atName))

        return d


class DataTransform(object):
    """ Factory for functional elements of the transformations between input data and
        and loadable data using specifications from the schema map definition.

        input string value -> (Null Handling -> cast type ->)

        input string iterable value ->
    """

    def __init__(self, tObj):
        #
        self.__wsPattern = re.compile(r"\s+", flags=re.UNICODE | re.MULTILINE)

        # SchemaDef Table Object -
        self.__tObj = tObj
        #

    def normalizeEnum(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        #origLength = len(trfTup.value)
        #if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
        #    return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        nVal = self.__tObj.normalizeEnum(trfTup.atId, trfTup.value)
        return TrfValue(nVal, trfTup.atId, trfTup.origLength, False)

    def castString(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        return TrfValue(trfTup.value, trfTup.atId, origLength, False)

    def castIterableString(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        vL = [v.strip() for v in trfTup.value.split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, origLength, False)

    def castInteger(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(str(trfTup.value))
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        return TrfValue(int(trfTup.value), trfTup.atId, origLength, False)

    def castIterableInteger(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(str(trfTup.value))
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        vL = [int(v.strip()) for v in trfTup.value.split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, origLength, False)

    def castFloat(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(str(trfTup.value))
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        return TrfValue(float(trfTup.value), trfTup.atId, origLength, False)

    def castIterableFloat(self, trfTup):
        """
            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(str(trfTup.value))
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        vL = [float(v.strip()) for v in trfTup.value.split(self.__tObj.getIterableSeparator(trfTup.atId))]
        return TrfValue(vL, trfTup.atId, origLength, False)

    def castDateToObj(self, trfTup):
        """ Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        return TrfValue(dateutil.parser.parse(tv), trfTup.atId, origLength, False)

    def castDateTimeToIsoDate(self, trfTup):
        """ Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        tS = dateutil.parser.parse(tv).replace(tzinfo=pytz.UTC).isoformat()

        return TrfValue(tS, trfTup.atId, origLength, False)

    def castDateToIsoDate(self, trfTup):
        """ Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) to a Python DateTime object -

            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        tv = trfTup.value.replace(":", " ", 1)
        tS = dateutil.parser.parse(tv).isoformat()

        return TrfValue(tS[:10], trfTup.atId, origLength, False)

    def castDateToString(self, trfTup):
        """ Cast the input date (optional time) string (yyyy-mm-dd:hh::mm:ss) as a string unchanged -

            Return:  TrfValue tuple
        """
        if trfTup.isNull:
            return trfTup
        origLength = len(trfTup.value)
        if ((origLength == 0) or (trfTup.value == '?') or (trfTup.value == '.')):
            return TrfValue(self.__tObj.getAppNullValue(trfTup.atId), trfTup.atId, origLength, True)
        return TrfValue(trfTup.value, trfTup.atId, origLength, False)

    def stripWhiteSpace(self, trfTup):
        """ Remove all white space from the input value.

            Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        value = self.__wsPattern.sub("", trfTup.value)
        return TrfValue(value, trfTup.atId, trfTup.origLength, False)

    def truncateString(self, trfTup):
        """ Truncate string value to maximum length setting.

            Return:  ReturnValue tuple
        """
        if trfTup.isNull:
            return trfTup
        return TrfValue(trfTup.value[:self.__tObj.getAttributeWidth(trfTup.atId)], trfTup.atId, trfTup.origLength, False)



##
# File:    ProvenanceUtil.py
# Author:  J. Westbrook
# Date:    23-Jun-2018
# Version: 0.001
#
# Updates:
#  24-Oct-2018  jdw revise api for new config organization
##
"""
 Utilities to access and update provenance details.

"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging

from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ProvenanceUtil(object):

    def __init__(self, cfgOb=None, **kwargs):
        self.__cfgOb = cfgOb
        self.__workPath = kwargs.get("workPath", None)

    def fetch(self, cfgSectionName='site_info'):
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            provLocator = self.__cfgOb.getPath('PROVENANCE_INFO_LOCATOR', sectionName=cfgSectionName)
            return mU.doImport(provLocator, format="json")
        except Exception as e:
            logger.exception("Failed retreiving provence for %s with %s" % (cfgSectionName, str(e)))
        return {}

    def update(self, provD, cfgSectionName='site_info'):
        ok = False
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            provLocator = self.__cfgOb.getPath('PROVENANCE_INFO_LOCATOR', sectionName=cfgSectionName)
            tD = mU.doImport(provLocator, format="json")
            tD.update(provD)
            ok = mU.doExport(provLocator, tD, format="json")
        except Exception as e:
            logger.exception("Failed storing provence for %s with %s" % (cfgSectionName, str(e)))
        return ok

    def store(self, provD, cfgSectionName='site_info'):
        ok = False
        try:
            mU = MarshalUtil(workPath=self.__workPath)
            provLocator = self.__cfgOb.getPath('PROVENANCE_INFO_LOCATOR', sectionName=cfgSectionName)
            ok = mU.doExport(provLocator, provD, format="json")
        except Exception as e:
            logger.exception("Failed storing provence for %s with %s" % (cfgSectionName, str(e)))
        return ok



##
# File:  ConnectionBase.py
# Date:  12-Mar-2018 J. Westbrook
#
# Update:
#    17-Mar-2018 jdw  add r/w sync controls - generalize auth to prefs
##
"""
Base class for managing database connection which handles application specific authentication.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import copy
import logging
import platform

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

try:
    # Python 3.x
    from urllib.parse import quote_plus
except ImportError:
    # Python 2.x
    from urllib import quote_plus

logger = logging.getLogger(__name__)


if platform.system() == "Linux":
    try:
        pass
    except Exception:
        logger.exception("Loading Linux feature failing")


class ConnectionBase(object):

    def __init__(self, verbose=False):
        self.__verbose = verbose
        #
        self.__infoD = {}
        self.__dbClient = None

        self.__databaseName = None
        self.__dbHost = None
        self.__dbUser = None
        self.__dbPw = None
        self.__dbSocket = None
        self.__dbPort = None
        self.__dbAdminDb = None
        self.__dbPort = None
        self.__defaultPort = 27017
        self.__dbServer = 'mongo'
        self.__resourceName = None

    def assignResource(self, resourceName=None):
        # implement in the derived class
        self._assignResource(resourceName)

    def _assignResource(self, resourceName):
        self.__resourceName = resourceName

    def getPreferences(self):
        return self.__infoD

    def setPreferences(self, infoD):
        try:
            self.__infoD = copy.deepcopy(infoD)
            self.__databaseName = self.__infoD.get("DB_NAME", None)
            self.__dbHost = self.__infoD.get("DB_HOST", 'localhost')
            self.__dbUser = self.__infoD.get("DB_USER", None)
            self.__dbPw = self.__infoD.get("DB_PW", None)
            self.__dbSocket = self.__infoD.get("DB_SOCKET", None)
            self.__dbServer = self.__infoD.get("DB_SERVER", "mongo")
            self.__dbAdminDb = self.__infoD.get("DB_ADMIN_DB_NAME", "admin")
            self.__writeConcern = self.__infoD.get("DB_WRITE_CONCERN", "majority")
            self.__readConcern = self.__infoD.get("DB_READ_CONCERN", "majority")
            self.__readPreference = self.__infoD.get("DB_READ_PREFERENCE", "nearest")
            self.__writeJournalOpt = self.__infoD.get("DB_WRITE_TO_JOURNAL", True)
            #
            port = self.__infoD.get("DB_PORT", self.__defaultPort)
            if port and len(str(port)) > 0:
                self.__dbPort = int(str(port))
        except Exception as e:
            logger.exception("Failing with %s" % str(e))

    def openConnection(self):
        """ Create a database connection and store a connection object.

            Returns True for success or False otherwise
        """
        #
        if self.__dbClient is not None:
            # Close an open connection -
            logger.warning("Closing an existing open connection.")
            self.closeConnection()

        try:
            if self.__dbUser and (len(self.__dbUser) > 0) and self.__dbPw and (len(self.__dbPw) > 0):
                uri = "mongodb://%s:%s@%s/%s" % (quote_plus(self.__dbUser), quote_plus(self.__dbPw), self.__dbHost, self.__dbAdminDb)
            else:
                uri = "mongodb://%s:%d" % (self.__dbHost, self.__dbPort)

            kw = {}
            kw['w'] = self.__writeConcern
            kw['j'] = True
            kw['appname'] = 'dbloader'
            kw['readConcernLevel'] = self.__readConcern
            kw['readPreference'] = self.__readPreference
            #
            # logger.debug("URI is %s" % uri)
            self.__dbClient = MongoClient(uri, **kw)
        except Exception as e:
            logger.error("Connection to resource %s failing with %s" % (self.__resourceName, str(e)))

        try:
            # The ismaster command is cheap and does not require auth.
            d = self.__dbClient.admin.command('ismaster')
            logger.debug("Server status: %r " % d)
            return True
        except ConnectionFailure:
            logger.exception("Connection failing to resource %s " % self.__resourceName)

        self.__dbClient = None

        return False

    def getClientConnection(self):
        """ Return an instance of a connected client.
        """
        return self.__dbClient

    def closeConnection(self):
        """ Close db session
        """
        if self.__dbClient is not None:
            self.__dbClient.close()
            self.__dbClient = None
            return True
        else:
            return False
